# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import inspect

import json
import netaddr
from neutron.db.sqlalchemyutils import paginate_query
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy import event
from sqlalchemy import func as sql_func
from sqlalchemy import and_, asc, desc, orm, or_, not_
from sqlalchemy.orm import class_mapper

from quark.db import models
from quark import network_strategy
from quark import protocols
from quark import tags

PORT_TAG_REGISTRY = tags.PORT_TAG_REGISTRY
STRATEGY = network_strategy.STRATEGY
LOG = logging.getLogger(__name__)
CONF = cfg.CONF


ONE = "one"
ALL = "all"

_NO_TENANT_MODELS = (
    models.PortIpAssociation,
    models.LockHolder,
    models.SegmentAllocationRange)


# NOTE(jkoelker) init event listener that will ensure id is filled in
#                on object creation (prior to commit).
def _perhaps_generate_id(target, args, kwargs):
    if hasattr(target, 'id') and target.id is None:
        target.id = uuidutils.generate_uuid()


# NOTE(jkoelker) Register the event on all models that have ids
for _name, klass in inspect.getmembers(models, inspect.isclass):
    if klass is models.HasId:
        continue

    if models.HasId in klass.mro():
        event.listen(klass, "init", _perhaps_generate_id)


def _listify(filters):
    for key in ["name", "network_id", "id", "device_id", "tenant_id",
                "subnet_id", "mac_address", "shared", "version", "segment_id",
                "device_owner", "ip_address", "used_by_tenant_id", "group_id"]:
        if key in filters:
            if not filters[key]:
                continue
            listified = filters[key]
            if not isinstance(listified, list):
                listified = [listified]
            filters[key] = listified


def _model_attrs(model):
    model_map = class_mapper(model)
    model_attrs = [x.key for x in model_map.column_attrs]
    if "_cidr" in model_attrs:
        model_attrs.append("cidr")
    if "_deallocated" in model_attrs:
        model_attrs.append("deallocated")
    return model_attrs


def _model_query(context, model, filters, fields=None):
    filters = filters or {}
    model_filters = []
    eq_filters = ["address", "cidr", "deallocated", "ip_version", "service",
                  "mac_address_range_id", "transaction_id", "lock_id",
                  "address_type", "completed"]
    in_filters = ["device_id", "device_owner", "group_id", "id", "mac_address",
                  "name", "network_id", "segment_id", "subnet_id",
                  "used_by_tenant_id", "version"]

    # Sanitize incoming filters to only attributes that exist in the model.
    # NOTE: Filters for unusable attributes are silently dropped here.
    # NOTE: When the filter key != attribute key, a conditional must be added
    #       here.
    model_attrs = _model_attrs(model)
    filters = {x: y for x, y in filters.items()
               if x in model_attrs or
               (x == "tenant_id" and model == models.IPAddress) or
               (x == "ip_address" and model == models.IPAddress) or
               (x == "reuse_after" and model in (models.IPAddress,
                                                 models.MacAddress))}

    # Inject the tenant id if none is set. We don't need unqualified queries.
    # This works even when a non-shared, other-tenant owned network is passed
    # in because the authZ checks that happen in Neutron above us yank it back
    # out of the result set.
    if not filters.get("tenant_id") and not context.is_admin:
        filters["tenant_id"] = [context.tenant_id]

    if model == models.SecurityGroupRule:
        sg_rule_attribs = ["direction", "port_range_max", "port_range_min"]
        eq_filters.extend(sg_rule_attribs)

    for key, value in filters.items():
        if key in in_filters:
            model_type = getattr(model, key)
            model_filters.append(model_type.in_(value))
        elif key in eq_filters:
            model_type = getattr(model, key)
            model_filters.append(model_type == value)
        elif key == "_deallocated":
            if value:
                model_filters.append(model._deallocated == 1)
            else:
                model_filters.append(model._deallocated != 1)
        elif key == "ethertype":
            etypes = []
            for etype in value:
                etypes.append(protocols.translate_ethertype(etype))
            model_filters.append(model.ethertype.in_(etypes))
        elif key == "ip_address":
            model_filters.append(model.address.in_(
                [ip.ipv6().value for ip in value]))
        elif key == 'protocol':
            pnums = []
            for version in (protocols.PROTOCOLS_V4, protocols.PROTOCOLS_V6):
                pnums.extend([y for x, y in version.items() if x in value])
            model_filters.append(model.protocol.in_(pnums))
        elif key == "reuse_after":
            reuse = (timeutils.utcnow() -
                     datetime.timedelta(seconds=value))
            # NOTE(asadoughi): should this allow for deallocated_at = null?
            model_filters.append(model.deallocated_at <= reuse)
        elif key == "port_id":
            if model == models.PortIpAssociation:
                model_filters.append(model.port_id == value)
        elif key == "tenant_id":
            if model == models.IPAddress:
                model_filters.append(model.used_by_tenant_id.in_(value))
            elif model in _NO_TENANT_MODELS:
                pass
            else:
                model_filters.append(model.tenant_id.in_(value))

    return model_filters


def scoped(f):
    def wrapped(*args, **kwargs):
        scope = None
        if "scope" in kwargs:
            scope = kwargs.pop("scope")
        if scope not in [None, ALL, ONE]:
            raise Exception("Invalid scope")
        _listify(kwargs)

        res = f(*args, **kwargs)
        if not res:
            return
        if "order_by" in kwargs:
            res = res.order_by(kwargs["order_by"])

        if scope == ALL:
            if isinstance(res, list):
                return res
            return res.all()
        elif scope == ONE:
            if isinstance(res, list):
                return res[0]
            return res.first()
        return res
    return wrapped


@scoped
def port_find(context, limit=None, sorts=None, marker_obj=None, fields=None,
              **filters):
    query = context.session.query(models.Port).options(
        orm.joinedload(models.Port.ip_addresses))
    model_filters = _model_query(context, models.Port, filters)
    if filters.get("ip_address_id"):
        model_filters.append(models.Port.ip_addresses.any(
            models.IPAddress.id.in_(filters["ip_address_id"])))

    if filters.get("device_id"):
        model_filters.append(models.Port.device_id.in_(filters["device_id"]))

    if filters.get("service"):
        model_filters.append(models.Port.associations.any(
            models.PortIpAssociation.service == filters["service"]))

    if "join_security_groups" in filters:
        query = query.options(orm.joinedload(models.Port.security_groups))

    if fields and "port_subnets" in fields:
        query = query.options(orm.joinedload("ip_addresses.subnet"))
        query = query.options(
            orm.joinedload("ip_addresses.subnet.dns_nameservers"))
        query = query.options(
            orm.joinedload("ip_addresses.subnet.routes"))
    return paginate_query(query.filter(*model_filters), models.Port, limit,
                          sorts, marker_obj)


@scoped
def port_find_by_ip_address(context, **filters):
    query = context.session.query(models.IPAddress).options(
        orm.joinedload(models.IPAddress.ports))
    model_filters = _model_query(context, models.IPAddress, filters)
    return query.filter(*model_filters)


def port_count_all(context, **filters):
    query = context.session.query(sql_func.count(models.Port.id))
    model_filters = _model_query(context, models.Port, filters)
    return query.filter(*model_filters).scalar()


def port_create(context, **port_dict):
    port = models.Port(tags=[])
    port.update(port_dict)
    port["tenant_id"] = context.tenant_id
    if "addresses" in port_dict:
        port["ip_addresses"].extend(port_dict["addresses"])
    PORT_TAG_REGISTRY.set_all(port, **port_dict)
    context.session.add(port)
    return port


@scoped
def ip_port_association_find(context, **filters):
    query = context.session.query(models.PortIpAssociation)
    model_filters = _model_query(context, models.PortIpAssociation, filters)
    return query.filter(*model_filters)


def port_disassociate_ip(context, ports, address):
    assocs_to_remove = [assoc for assoc in address.associations
                        if assoc.port in ports]
    for assoc in assocs_to_remove:
        context.session.delete(assoc)
        # NOTE(thomasem): Need to update in-session model for caller.
        address.associations.remove(assoc)
    context.session.add(address)
    return address


def port_associate_ip(context, ports, address, enable_port=None):
    for port in ports:
        assoc = models.PortIpAssociation()
        assoc.port = port
        assoc.ip_address = address
        assoc.enabled = port.id in enable_port if enable_port else False
        context.session.add(assoc)
    return address


def get_ports_for_address(address):
    ports = []
    for assoc in address.associations:
        ports.append(assoc.port)
    return ports


def update_port_associations_for_ip(context, ports, address):
    assoc_ports = set(get_ports_for_address(address))
    new_ports = set(ports)
    new_address = port_associate_ip(context, new_ports - assoc_ports,
                                    address)
    return port_disassociate_ip(context,
                                assoc_ports - new_ports, new_address)


def port_update(context, port, **kwargs):
    if "addresses" in kwargs:
        port["ip_addresses"] = kwargs.pop("addresses")
    PORT_TAG_REGISTRY.set_all(port, **kwargs)
    port.update(kwargs)
    context.session.add(port)
    return port


def port_delete(context, port):
    context.session.delete(port)


def ip_address_update(context, address, **kwargs):
    address.update(kwargs)
    context.session.add(address)
    return address


def ip_address_create(context, **address_dict):
    ip_address = models.IPAddress()
    address = address_dict.pop("address")
    ip_address.update(address_dict)
    ip_address["address"] = int(address.ipv6())
    ip_address["address_readable"] = str(address)
    ip_address["used_by_tenant_id"] = context.tenant_id
    ip_address["_deallocated"] = 0
    ip_address["allocated_at"] = timeutils.utcnow()
    context.session.add(ip_address)
    return ip_address


def ip_address_delete(context, addr):
    context.session.delete(addr)


def ip_address_deallocate(context, address, **kwargs):
    kwargs["_deallocated"] = 1
    kwargs["deallocated_at"] = timeutils.utcnow()
    return ip_address_update(context, address, **kwargs)


@scoped
def ip_address_find(context, lock_mode=False, **filters):
    query = context.session.query(models.IPAddress)

    if lock_mode:
        query = query.with_lockmode("update")

    # NOTE(asadoughi): NCP-1677: Maps string or integer into integer value
    if filters.get("address"):
        filters["address"] = int(netaddr.IPAddress(filters["address"]).ipv6())

    model_filters = _model_query(context, models.IPAddress, filters)

    if filters.get("device_id"):
        model_filters.append(models.IPAddress.ports.any(
            models.Port.device_id.in_(filters["device_id"])))

    if filters.get("service"):
        model_filters.append(models.IPAddress.associations.any(
            models.PortIpAssociation.service == filters["service"]))

    if filters.get("port_id"):
        model_filters.append(models.IPAddress.ports.any(
            models.Port.id == filters['port_id']))

    return query.filter(*model_filters)


def ip_address_count_all(context, filters):
    query = context.session.query(sql_func.count(models.IPAddress.id))
    model_filters = _model_query(context, models.IPAddress, filters)
    return query.filter(*model_filters).scalar()


@scoped
def ip_address_reallocate(context, update_kwargs, **filters):
    LOG.debug("ip_address_reallocate %s", filters)
    query = context.session.query(models.IPAddress)
    model_filters = _model_query(context, models.IPAddress, filters)
    query = query.filter(*model_filters)
    row_count = query.update(update_kwargs,
                             update_args={"mysql_limit": 1},
                             synchronize_session=False)
    return row_count == 1


def ip_address_reallocate_find(context, transaction_id):
    address = ip_address_find(context, transaction_id=transaction_id,
                              scope=ONE)
    if not address:
        LOG.warn("Couldn't find IP address with transaction_id %s",
                 transaction_id)
        return

    LOG.info("Potentially reallocatable IP found: "
             "{0}".format(address["address_readable"]))
    subnet = address.get('subnet')
    if not subnet:
        LOG.debug("No subnet associated with address")
        return
    if subnet["do_not_use"]:
        LOG.debug("Subnet marked as do_not_use")
        return

    addr = netaddr.IPAddress(int(address["address"]))
    if address["subnet"]["ip_version"] == 4:
        addr = addr.ipv4()
    else:
        addr = addr.ipv6()

    if subnet and subnet["ip_policy"]:
        policy = subnet["ip_policy"].get_cidrs_ip_set()
    else:
        policy = netaddr.IPSet([])

    if policy is not None and addr in policy:
        LOG.info("Deleting Address {0} due to policy "
                 "violation".format(
                     address["address_readable"]))
        context.session.delete(address)
        return

    # TODO(amir): performance test replacing this with SQL in
    #             ip_address_reallocate's UPDATE statement
    cidr = netaddr.IPNetwork(address["subnet"]["cidr"])
    if addr not in cidr:
        LOG.info("Address {0} isn't in the subnet "
                 "it claims to be in".format(
                     address["address_readable"]))
        context.session.delete(address)
        return

    return address


@scoped
def mac_address_find(context, lock_mode=False, **filters):
    query = context.session.query(models.MacAddress)
    if lock_mode:
        query = query.with_lockmode("update")
    model_filters = _model_query(context, models.MacAddress, filters)
    return query.filter(*model_filters)


def mac_address_delete(context, mac_address):
    context.session.delete(mac_address)


@scoped
def mac_address_reallocate(context, update_kwargs, **filters):
    LOG.debug("mac_address_reallocate %s", filters)
    query = context.session.query(models.MacAddress)
    model_filters = _model_query(context, models.MacAddress, filters)
    query = query.filter(*model_filters)
    row_count = query.update(update_kwargs, update_args={"mysql_limit": 1},
                             synchronize_session=False)
    return row_count == 1


def mac_address_reallocate_find(context, transaction_id):
    mac = mac_address_find(context, transaction_id=transaction_id,
                           scope=ONE)
    if not mac:
        LOG.warn("Couldn't find MAC address with transaction_id %s",
                 transaction_id)
        return

    # NOTE(mdietz): This is a HACK. Please see RM11043 for details
    if mac["mac_address_range"] and mac["mac_address_range"]["do_not_use"]:
        mac_address_delete(context, mac)
        LOG.debug("Found a deallocated MAC in a do_not_use"
                  " mac_address_range and deleted it. "
                  "Retrying...")
        return

    return mac


def mac_address_range_find_allocation_counts(context, address=None,
                                             use_forbidden_mac_range=False):
    count = sql_func.count(models.MacAddress.address)
    query = context.session.query(models.MacAddressRange,
                                  count.label("count")).with_lockmode("update")
    query = query.outerjoin(models.MacAddress)
    query = query.group_by(models.MacAddressRange.id)
    query = query.order_by(desc(count))
    if address:
        query = query.filter(models.MacAddressRange.last_address >= address)
        query = query.filter(models.MacAddressRange.first_address <= address)
    query = query.filter(models.MacAddressRange.next_auto_assign_mac != -1)
    if not use_forbidden_mac_range:
        query = query.filter(models.MacAddressRange.do_not_use == '0')  # noqa
    query = query.limit(1)
    return query.first()


@scoped
def mac_address_range_find(context, **filters):
    query = context.session.query(models.MacAddressRange)
    model_filters = _model_query(context, models.MacAddressRange, filters)
    return query.filter(*model_filters)


def mac_address_range_create(context, **range_dict):
    new_range = models.MacAddressRange()
    new_range.update(range_dict)
    context.session.add(new_range)
    return new_range


def mac_address_range_delete(context, mac_address_range):
    context.session.delete(mac_address_range)


def mac_address_range_update(context, mac_range, **kwargs):
    mac_range.update(kwargs)
    context.session.add(mac_range)
    return mac_range


def mac_range_update_next_auto_assign_mac(context, mac_range):
    query = context.session.query(models.MacAddressRange)
    query = query.filter(models.MacAddressRange.id == mac_range["id"])
    query = query.filter(models.MacAddressRange.next_auto_assign_mac != -1)

    # For details on synchronize_session, see:
    # http://docs.sqlalchemy.org/en/rel_0_8/orm/query.html
    query = query.update(
        {"next_auto_assign_mac":
         models.MacAddressRange.next_auto_assign_mac + 1},
        synchronize_session=False)

    # Returns a count of the rows matched in the update
    return query


def mac_range_update_set_full(context, mac_range):
    query = context.session.query(models.MacAddressRange)
    query = query.filter_by(id=mac_range["id"])
    query = query.filter(models.MacAddressRange.next_auto_assign_mac != -1)

    # For details on synchronize_session, see:
    # http://docs.sqlalchemy.org/en/rel_0_8/orm/query.html
    query = query.update(
        {"next_auto_assign_mac": -1},
        synchronize_session=False)

    # Returns a count of the rows matched in the update
    return query


def mac_address_update(context, mac, **kwargs):
    mac.update(kwargs)
    context.session.add(mac)
    return mac


def mac_address_create(context, **mac_dict):
    mac_address = models.MacAddress()
    mac_address.update(mac_dict)
    mac_address["tenant_id"] = context.tenant_id
    mac_address["deallocated"] = False
    mac_address["deallocated_at"] = None
    context.session.add(mac_address)
    return mac_address


INVERT_DEFAULTS = 'invert_defaults'


@scoped
def network_find(context, limit=None, sorts=None, marker=None,
                 page_reverse=False, fields=None, **filters):
    ids = []
    defaults = []
    provider_query = False
    if "id" in filters:
        ids, defaults = STRATEGY.split_network_ids(filters["id"])
        if not ids and defaults and "shared" not in filters:
            provider_query = True
        if ids:
            filters["id"] = ids
        else:
            filters.pop("id")

    if "shared" in filters:
        defaults = STRATEGY.get_provider_networks()
        if True in filters["shared"]:
            if ids:
                defaults = [net for net in ids if net in defaults]
                filters.pop("id")
            if not defaults:
                return []
        else:
            defaults.insert(0, INVERT_DEFAULTS)
        filters.pop("shared")
    return _network_find(context, limit, sorts, marker, page_reverse, fields,
                         defaults=defaults, provider_query=provider_query,
                         **filters)


def _network_find(context, limit, sorts, marker, page_reverse, fields,
                  defaults=None, provider_query=False, **filters):
    query = context.session.query(models.Network)
    model_filters = _model_query(context, models.Network, filters, query)

    if defaults:
        invert_defaults = False
        if INVERT_DEFAULTS in defaults:
            invert_defaults = True
            defaults.pop(0)
        if filters and invert_defaults:
            query = query.filter(and_(not_(models.Network.id.in_(defaults)),
                                      and_(*model_filters)))
        elif not provider_query and filters and not invert_defaults:
            query = query.filter(or_(models.Network.id.in_(defaults),
                                     and_(*model_filters)))

        elif not invert_defaults:
            query = query.filter(models.Network.id.in_(defaults))
    else:
        query = query.filter(*model_filters)

    if "join_subnets" in filters:
        query = query.options(orm.joinedload(models.Network.subnets))

    return paginate_query(query, models.Network, limit, sorts, marker)


def network_create(context, **network):
    new_net = models.Network()
    new_net.update(network)
    context.session.add(new_net)
    return new_net


def network_update(context, network, **kwargs):
    network.update(kwargs)
    context.session.add(network)
    return network


def network_count_all(context):
    query = context.session.query(sql_func.count(models.Network.id))
    return query.filter(
        models.Network.tenant_id == context.tenant_id).scalar()


def network_delete(context, network):
    context.session.delete(network)


def _subnet_find_ordered_by_used_ips(context, net_id, lock_subnets=True,
                                     order="asc", unused=False, **filters):
    count = sql_func.count(models.IPAddress.address).label("count")
    size = (models.Subnet.last_ip - models.Subnet.first_ip)
    query = context.session.query(models.Subnet, count)
    if lock_subnets:
        query = query.with_lockmode("update")
    query = query.filter_by(do_not_use=False)
    query = query.outerjoin(models.Subnet.generated_ips)
    query = query.group_by(models.Subnet.id)

    query = query.order_by(asc(models.Subnet.ip_version))

    if unused:  # find unused subnets
        query = query.having(count == 0)
    else:  # otherwise, order used subnets
        if order == "desc":
            query = query.order_by(desc(size - count))
        else:  # default acending
            query = query.order_by(asc(size - count))

    query = query.filter(models.Subnet.network_id == net_id)
    if "ip_version" in filters:
        query = query.filter(models.Subnet.ip_version == filters["ip_version"])
    if "segment_id" in filters and filters["segment_id"]:
        query = query.filter(models.Subnet.segment_id == filters["segment_id"])
    query = query.filter(models.Subnet.next_auto_assign_ip != -1)

    if "subnet_id" in filters and filters["subnet_id"]:
        query = query.filter(models.Subnet.id.in_(filters["subnet_id"]))
    return query


def subnet_find_ordered_by_most_full(context, net_id, lock_subnets=True,
                                     **filters):
    return _subnet_find_ordered_by_used_ips(
        context, net_id, lock_subnets=lock_subnets, order="asc", **filters)


def subnet_find_ordered_by_least_full(context, net_id, lock_subnets=True,
                                      **filters):
    return _subnet_find_ordered_by_used_ips(
        context, net_id, lock_subnets=lock_subnets, order="desc", **filters)


def subnet_find_unused(context, net_id, lock_subnets=True,
                       **filters):
    return _subnet_find_ordered_by_used_ips(
        context, net_id, lock_subnets=lock_subnets, unused=True, **filters)


def subnet_update_next_auto_assign_ip(context, subnet):
    query = context.session.query(models.Subnet)
    query = query.filter(models.Subnet.id == subnet["id"])
    query = query.filter(models.Subnet.next_auto_assign_ip != -1)

    # For details on synchronize_session, see:
    # http://docs.sqlalchemy.org/en/rel_0_8/orm/query.html
    query = query.update(
        {"next_auto_assign_ip":
         models.Subnet.next_auto_assign_ip + 1},
        synchronize_session=False)

    # Returns a count of the rows matched in the update
    return query


def subnet_update_set_full(context, subnet):
    query = context.session.query(models.Subnet)
    query = query.filter_by(id=subnet["id"])
    query = query.filter(models.Subnet.next_auto_assign_ip != -1)

    # For details on synchronize_session, see:
    # http://docs.sqlalchemy.org/en/rel_0_8/orm/query.html
    query = query.update(
        {"next_auto_assign_ip": -1},
        synchronize_session=False)

    # Returns a count of the rows matched in the update
    return query


def subnet_update_set_alloc_pool_cache(context, subnet, cache_data=None):
    if cache_data is not None:
        cache_data = json.dumps(cache_data)
    update_kwargs = {"_allocation_pool_cache": cache_data}
    query = context.session.query(models.Subnet)
    query = query.filter(models.Subnet.id == subnet.id)
    row_count = query.update(update_kwargs,
                             update_args={"mysql_limit": 1})
    LOG.debug("Setting alloc pool cache to %s (row_count: %s)" % (
        cache_data, row_count))
    return subnet


@scoped
def subnet_find(context, limit=None, page_reverse=False, sorts=None,
                marker_obj=None, fields=None, **filters):
    ids = []
    defaults = []
    provider_query = False
    if "id" in filters:
        ids, defaults = STRATEGY.split_subnet_ids(filters["id"])
        if not ids and defaults and "shared" not in filters:
            provider_query = True
        if ids:
            filters["id"] = ids
        else:
            filters.pop("id")

    if "shared" in filters:
        defaults = STRATEGY.get_provider_subnets()
        if True in filters["shared"]:
            if ids:
                defaults = [sub for sub in ids if sub in defaults]
                filters.pop("id")
            if not defaults:
                return []
        else:
            defaults.insert(0, INVERT_DEFAULTS)
        filters.pop("shared")
    return _subnet_find(context, limit, sorts, marker_obj, page_reverse,
                        fields, defaults=defaults,
                        provider_query=provider_query, **filters)


def _subnet_find(context, limit, sorts, marker, page_reverse, fields,
                 defaults=None, provider_query=False, **filters):
    query = context.session.query(models.Subnet)
    model_filters = _model_query(context, models.Subnet, filters, query)

    if defaults:
        invert_defaults = False
        if INVERT_DEFAULTS in defaults:
            invert_defaults = True
            defaults.pop(0)
        if filters and invert_defaults:
            query = query.filter(and_(not_(models.Subnet.id.in_(defaults)),
                                      and_(*model_filters)))
        elif not provider_query and filters and not invert_defaults:
            query = query.filter(or_(models.Subnet.id.in_(defaults),
                                     and_(*model_filters)))

        elif not invert_defaults:
            query = query.filter(models.Subnet.id.in_(defaults))
    else:
        query = query.filter(*model_filters)

    if "join_dns" in filters:
        query = query.options(orm.joinedload(models.Subnet.dns_nameservers))

    if "join_routes" in filters:
        query = query.options(orm.joinedload(models.Subnet.routes))

    if "join_pool" in filters:
        query = query.options(orm.undefer('_allocation_pool_cache'))

    return paginate_query(query, models.Subnet, limit, sorts, marker)


def subnet_count_all(context, **filters):
    query = context.session.query(sql_func.count(models.Subnet.id))
    if filters.get("network_id"):
        query = query.filter(
            models.Subnet.network_id == filters["network_id"])
    query.filter(models.Subnet.tenant_id == context.tenant_id)
    return query.scalar()


def subnet_delete(context, subnet):
    context.session.delete(subnet)


def subnet_create(context, **subnet_dict):
    subnet = models.Subnet()
    subnet.update(subnet_dict)
    subnet["tenant_id"] = context.tenant_id
    context.session.add(subnet)
    return subnet


def subnet_update(context, subnet, **kwargs):
    subnet.update(kwargs)
    context.session.add(subnet)
    return subnet


@scoped
def route_find(context, fields=None, **filters):
    query = context.session.query(models.Route)
    model_filters = _model_query(context, models.Route, filters)
    return query.filter(*model_filters)


def route_create(context, **route_dict):
    new_route = models.Route()
    new_route.update(route_dict)
    new_route["tenant_id"] = context.tenant_id
    context.session.add(new_route)
    return new_route


def route_update(context, route, **kwargs):
    route.update(kwargs)
    context.session.add(route)
    return route


def route_delete(context, route):
    context.session.delete(route)


def dns_create(context, **dns_dict):
    dns_nameserver = models.DNSNameserver()
    ip = dns_dict.pop("ip")
    dns_nameserver.update(dns_dict)
    dns_nameserver["ip"] = int(ip)
    dns_nameserver["tenant_id"] = context.tenant_id
    context.session.add(dns_nameserver)
    return dns_nameserver


def dns_delete(context, dns):
    context.session.delete(dns)


@scoped
def security_group_find(context, **filters):
    query = context.session.query(models.SecurityGroup).options(
        orm.joinedload(models.SecurityGroup.rules))
    model_filters = _model_query(context, models.SecurityGroup, filters)
    return query.filter(*model_filters)


@scoped
def security_group_count(context, **filters):
    query = context.session.query(sql_func.count(models.SecurityGroup.id))
    model_filters = _model_query(context, models.SecurityGroup, filters)
    return query.filter(*model_filters).scalar()


@scoped
def ports_with_security_groups_find(context):
    query = context.session.query(models.Port)
    query = query.join(models.Port.security_groups)
    query = query.options(orm.contains_eager(models.Port.security_groups))
    return query


@scoped
def ports_with_security_groups_count(context):
    query = context.session.query(
        sql_func.count(models.port_group_association_table.c.port_id))
    return query.scalar()


def security_group_create(context, **sec_group_dict):
    new_group = models.SecurityGroup()
    new_group.update(sec_group_dict)
    new_group["tenant_id"] = context.tenant_id
    context.session.add(new_group)
    return new_group


def security_group_update(context, group, **kwargs):
    group.update(kwargs)
    context.session.add(group)
    return group


def security_group_delete(context, group):
    context.session.delete(group)


@scoped
def security_group_rule_find(context, **filters):
    query = context.session.query(models.SecurityGroupRule)
    model_filters = _model_query(context, models.SecurityGroupRule, filters)
    return query.filter(*model_filters)


def security_group_rule_create(context, **rule_dict):
    new_rule = models.SecurityGroupRule()
    new_rule.update(rule_dict)
    new_rule.group_id = rule_dict['security_group_id']
    new_rule.tenant_id = rule_dict['tenant_id']
    context.session.add(new_rule)
    return new_rule


def security_group_rule_delete(context, rule):
    context.session.delete(rule)


def ip_policy_create(context, **ip_policy_dict):
    new_policy = models.IPPolicy()
    exclude = ip_policy_dict.pop("exclude")
    ip_set = netaddr.IPSet()
    for excluded_cidr in exclude:
        cidr_net = netaddr.IPNetwork(excluded_cidr).ipv6()
        new_policy["exclude"].append(
            models.IPPolicyCIDR(cidr=excluded_cidr,
                                first_ip=cidr_net.first,
                                last_ip=cidr_net.last))
        ip_set.add(excluded_cidr)
    ip_policy_dict["size"] = ip_set.size
    new_policy.update(ip_policy_dict)
    new_policy["tenant_id"] = context.tenant_id
    context.session.add(new_policy)
    return new_policy


@scoped
def ip_policy_find(context, **filters):
    query = context.session.query(models.IPPolicy)
    model_filters = _model_query(context, models.IPPolicy, filters)
    return query.filter(*model_filters)


def ip_policy_update(context, ip_policy, **ip_policy_dict):
    exclude = ip_policy_dict.pop("exclude", [])
    if exclude:
        ip_policy["exclude"] = []
        ip_set = netaddr.IPSet()
        for excluded_cidr in exclude:
            cidr_net = netaddr.IPNetwork(excluded_cidr).ipv6()
            ip_policy["exclude"].append(
                models.IPPolicyCIDR(cidr=excluded_cidr,
                                    first_ip=cidr_net.first,
                                    last_ip=cidr_net.last))
            ip_set.add(excluded_cidr)
        ip_policy_dict["size"] = ip_set.size

    ip_policy.update(ip_policy_dict)
    context.session.add(ip_policy)
    return ip_policy


def ip_policy_delete(context, ip_policy):
    context.session.delete(ip_policy)


def transaction_create(context):
    transaction = models.Transaction()
    context.session.add(transaction)
    return transaction


@scoped
def floating_ip_find(context, lock_mode=False, limit=None, sorts=None,
                     marker=None, page_reverse=False, fields=None, **filters):
    query = context.session.query(models.IPAddress)

    if lock_mode:
        query = query.with_lockmode("update")

    model_filters = _model_query(context, models.IPAddress, filters)

    if filters.get("port_id"):
        model_filters.append(models.IPAddress.ports.any(
            models.Port.id == filters['port_id']))

    if filters.get("transaction_id"):
        model_filters.append(
            models.IPAddress.transaction_id == filters['transaction_id'])

    return paginate_query(query.filter(*model_filters), models.IPAddress,
                          limit, sorts, marker)


def floating_ip_associate_fixed_ip(context, floating_ip, fixed_ip):
    floating_ip.fixed_ips.append(fixed_ip)
    return floating_ip


def floating_ip_disassociate_fixed_ip(context, floating_ip, fixed_ip):
    found_index = None
    for index, flip_fixed_ip in enumerate(floating_ip.fixed_ips):
        if flip_fixed_ip.id == fixed_ip.id:
            found_index = index
            break
    floating_ip.fixed_ips.pop(found_index)
    return floating_ip


def floating_ip_disassociate_all_fixed_ips(context, floating_ip):
    floating_ip.fixed_ips = []
    return floating_ip


@scoped
def lock_holder_find(context, **filters):
    query = context.session.query(models.LockHolder)
    model_filters = _model_query(context, models.LockHolder, filters)
    return query.filter(*model_filters)


def _lock_create(context, target, **kwargs):
    with context.session.begin():
        result = context.session.execute(
            models.Lock.__table__.insert(),
            {"type": kwargs["type"]})
        lock_id = result.lastrowid

        target_model = target.__class__
        row_count = context.session.query(target_model).filter(
            and_(target_model.id == target.id, target_model.lock_id.is_(None))
        ).update(dict(lock_id=lock_id))

    if row_count == 0:
        context.session.query(models.Lock).filter(
            models.Lock.id == lock_id).delete()
        return None
    return lock_id


def lock_holder_create(context, target, **kwargs):
    lock_id = target.lock_id
    if not lock_id:
        lock_id = _lock_create(context, target, **kwargs)
        if not lock_id:
            # Failed to concurrently create lock; invoker must retry
            return
    lock_holder = models.LockHolder(lock_id=lock_id,
                                    name=kwargs["name"])
    context.session.add(lock_holder)
    return lock_holder


def _lock_delete(context, target):
    with context.session.begin():
        lock_id = target.lock_id
        lock_holder = lock_holder_find(context, lock_id=lock_id, scope=ONE)
        if not lock_holder:
            target.lock_id = None
            context.session.add(target)
            context.session.query(
                models.Lock).filter(
                    models.Lock.id == lock_id).delete()


def lock_holder_delete(context, target, lock_holder):
    context.session.delete(lock_holder)
    _lock_delete(context, target)


@scoped
def segment_allocation_range_find(context, lock_mode=False, **filters):
    query = context.session.query(models.SegmentAllocationRange)
    if lock_mode:
        query = query.with_lockmode("update")

    model_filters = _model_query(
        context, models.SegmentAllocationRange, filters)

    query = query.filter(*model_filters)
    return query


def segment_allocation_find(context, lock_mode=False, **filters):
    """Query for segment allocations."""
    range_ids = filters.pop("segment_allocation_range_ids", None)

    query = context.session.query(models.SegmentAllocation)
    if lock_mode:
        query = query.with_lockmode("update")

    query = query.filter_by(**filters)

    # Optionally filter by given list of range ids
    if range_ids:
        query.filter(
            models.SegmentAllocation.segment_allocation_range_id.in_(
                range_ids))
    return query


def segment_allocation_update(context, sa, **sa_dict):
    sa.update(sa_dict)
    context.session.add(sa)
    return sa


def segment_allocation_range_populate_bulk(context, sa_dicts):
    """Bulk-insert deallocated segment allocations.

    NOTE(morgabra): This is quite performant when populating large ranges,
    but you don't get any ORM conveniences or protections here.
    """
    context.session.bulk_insert_mappings(
        models.SegmentAllocation,
        sa_dicts
    )


def segment_allocation_range_create(context, **sa_range_dict):
    new_range = models.SegmentAllocationRange()
    new_range.update(sa_range_dict)
    context.session.add(new_range)
    return new_range


def segment_allocation_range_delete(context, sa_range):
    context.session.delete(sa_range)


@scoped
def async_transaction_find(context, lock_mode=False, **filters):
    query = context.session.query(models.AsyncTransactions)
    if lock_mode:
        query = query.with_lockmode("update")

    model_filters = _model_query(
        context, models.AsyncTransactions, filters)

    query = query.filter(*model_filters)
    return query


def async_transaction_create(context, **transaction_dict):
    tx = models.AsyncTransactions()
    tx.update(transaction_dict)
    context.session.add(tx)
    return tx


def async_transaction_update(context, transaction, **kwargs):
    transaction.update(kwargs)
    context.session.add(transaction)
    return transaction


def async_transaction_delete(context, transaction):
    context.session.delete(transaction)



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

quark_router_opts = [
    cfg.StrOpt('floating_ip_router_id',
               default='00000000-0000-0000-0000-000000000000',
               help=_('floating ips default public router id'))
]

CONF.register_opts(quark_router_opts, "QUARK")


def get_router(context, id, fields):
    LOG.info("get_router %s for tenant %s fields %s" %
             (id, context.tenant_id, fields))
    if id != CONF.QUARK.floating_ip_router_id:
        return None

    return _get_floating_ip_default_router(context.tenant_id)


def get_routers(context, filters=None, fields=None, sorts=None, limit=None,
                marker=None, page_reverse=False):
    LOG.info("get_routers for tenant %s filters %s fields %s" %
             (context.tenant_id, filters, fields))
    return [_get_floating_ip_default_router(context.tenant_id)]


def _get_floating_ip_default_router(tenant_id):
    return {"id": CONF.QUARK.floating_ip_router_id,
            "status": "ACTIVE",
            "tenant_id": tenant_id,
            "name": "Floating IP Public Router",
            "admin_state_up": True}



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging

from quark import allocation_pool
from quark.db import api as db_api
from quark import exceptions as q_exc
from quark import plugin_views as v

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def _validate_policy_with_routes(context, policies, subnets):
    pools = {}
    policy_networks = [netaddr.IPNetwork(p) for p in policies]
    for subnet in subnets:
        pool = allocation_pool.AllocationPools(subnet["cidr"],
                                               policies=policy_networks)
        pools[subnet["id"]] = pool

    subnet_ids = [subnet["id"] for subnet in subnets]

    routes = db_api.route_find(context, subnet_id=subnet_ids)
    for route in routes:
        subnet_pool = pools[route["subnet_id"]]
        subnet_pool.validate_gateway_excluded(route["gateway"])


def create_ip_policy(context, ip_policy):
    LOG.info("create_ip_policy for tenant %s" % context.tenant_id)

    ipp = ip_policy['ip_policy']

    if not ipp.get("exclude"):
        raise n_exc.BadRequest(resource="ip_policy",
                               msg="Empty ip_policy.exclude")

    network_ids = ipp.get("network_ids")
    subnet_ids = ipp.get("subnet_ids")

    if subnet_ids and network_ids:
        raise n_exc.BadRequest(
            resource="ip_policy",
            msg="network_ids and subnet_ids specified. only one allowed")

    if not subnet_ids and not network_ids:
        raise n_exc.BadRequest(
            resource="ip_policy",
            msg="network_ids or subnet_ids not specified")

    with context.session.begin():
        if subnet_ids:
            subnets = db_api.subnet_find(
                context, id=subnet_ids, scope=db_api.ALL)
            if not subnets:
                raise n_exc.SubnetNotFound(subnet_id=subnet_ids)
            _check_for_pre_existing_policies_in(subnets)
            ensure_default_policy(ipp["exclude"], subnets)
            _validate_cidrs_fit_into_subnets(ipp["exclude"], subnets)
            ipp.pop("subnet_ids")
            ipp["subnets"] = subnets

        if network_ids:
            nets = db_api.network_find(
                context, id=network_ids, scope=db_api.ALL)
            if not nets:
                raise n_exc.NetworkNotFound(net_id=network_ids)
            _check_for_pre_existing_policies_in(nets)
            subnets = [subnet for net in nets
                       for subnet in net.get("subnets", [])]
            ensure_default_policy(ipp["exclude"], subnets)
            _validate_cidrs_fit_into_subnets(ipp["exclude"], subnets)
            ipp.pop("network_ids")
            ipp["networks"] = nets

        ip_policy = db_api.ip_policy_create(context, **ipp)
    return v._make_ip_policy_dict(ip_policy)


def _check_for_pre_existing_policies_in(models):
    models_with_existing_policies = [model for model in models
                                     if model.get('ip_policy', None)]
    if models_with_existing_policies:
        first_model = models_with_existing_policies[0]
        raise q_exc.IPPolicyAlreadyExists(
            id=first_model['ip_policy']['id'],
            n_id=first_model['id'])


def get_ip_policy(context, id):
    LOG.info("get_ip_policy %s for tenant %s" % (id, context.tenant_id))
    ipp = db_api.ip_policy_find(context, id=id, scope=db_api.ONE)
    if not ipp:
        raise q_exc.IPPolicyNotFound(id=id)
    return v._make_ip_policy_dict(ipp)


def get_ip_policies(context, **filters):
    LOG.info("get_ip_policies for tenant %s" % (context.tenant_id))
    ipps = db_api.ip_policy_find(context, scope=db_api.ALL, **filters)
    return [v._make_ip_policy_dict(ipp) for ipp in ipps]


def update_ip_policy(context, id, ip_policy):
    LOG.info("update_ip_policy for tenant %s" % context.tenant_id)

    ipp = ip_policy["ip_policy"]

    with context.session.begin():
        ipp_db = db_api.ip_policy_find(context, id=id, scope=db_api.ONE)
        if not ipp_db:
            raise q_exc.IPPolicyNotFound(id=id)

        ip_policy_cidrs = ipp.get("exclude")
        network_ids = ipp.get("network_ids")
        subnet_ids = ipp.get("subnet_ids")

        if subnet_ids and network_ids:
            raise n_exc.BadRequest(
                resource="ip_policy",
                msg="network_ids and subnet_ids specified. only one allowed")

        models = []
        all_subnets = []
        if subnet_ids:
            for subnet in ipp_db["subnets"]:
                subnet["ip_policy"] = None
            subnets = db_api.subnet_find(
                context, id=subnet_ids, scope=db_api.ALL)
            if len(subnets) != len(subnet_ids):
                raise n_exc.SubnetNotFound(subnet_id=subnet_ids)
            if ip_policy_cidrs is not None:
                ensure_default_policy(ip_policy_cidrs, subnets)
                _validate_cidrs_fit_into_subnets(ip_policy_cidrs, subnets)
            all_subnets.extend(subnets)
            models.extend(subnets)

        if network_ids:
            for network in ipp_db["networks"]:
                network["ip_policy"] = None
            nets = db_api.network_find(context, id=network_ids,
                                       scope=db_api.ALL)
            if len(nets) != len(network_ids):
                raise n_exc.NetworkNotFound(net_id=network_ids)
            subnets = [subnet for net in nets
                       for subnet in net.get("subnets", [])]
            if ip_policy_cidrs is not None:
                ensure_default_policy(ip_policy_cidrs, subnets)
                _validate_cidrs_fit_into_subnets(ip_policy_cidrs, subnets)
            all_subnets.extend(subnets)
            models.extend(nets)

        if not subnet_ids and not network_ids and ip_policy_cidrs is not None:
            ensure_default_policy(ip_policy_cidrs, ipp_db["subnets"])
            _validate_cidrs_fit_into_subnets(
                ip_policy_cidrs, ipp_db["subnets"])

        for model in models:
            if model["ip_policy"]:
                raise q_exc.IPPolicyAlreadyExists(
                    id=model["ip_policy"]["id"], n_id=model["id"])
            model["ip_policy"] = ipp_db

        if ip_policy_cidrs:
            _validate_policy_with_routes(context, ip_policy_cidrs, all_subnets)
        ipp_db = db_api.ip_policy_update(context, ipp_db, **ipp)
    return v._make_ip_policy_dict(ipp_db)


def delete_ip_policy(context, id):
    LOG.info("delete_ip_policy %s for tenant %s" % (id, context.tenant_id))
    with context.session.begin():
        ipp = db_api.ip_policy_find(context, id=id, scope=db_api.ONE)
        if not ipp:
            raise q_exc.IPPolicyNotFound(id=id)
        if ipp["networks"] or ipp["subnets"]:
            raise q_exc.IPPolicyInUse(id=id)
        db_api.ip_policy_delete(context, ipp)


def _validate_cidrs_fit_into_subnets(cidrs, subnets):
    LOG.info("validate_cidrs_all_fit_into_subnets with CIDRs (%s) "
             "and subnets (%s)" % (cidrs, subnets))
    for cidr in cidrs:
        cidr = netaddr.IPNetwork(cidr)
        for subnet in subnets:
            subnet_cidr = netaddr.IPNetwork(subnet["cidr"])
            if cidr.version == subnet_cidr.version and cidr not in subnet_cidr:
                raise n_exc.BadRequest(
                    resource="ip_policy",
                    msg="CIDR %s not in subnet CIDR %s"
                    % (cidr, subnet_cidr))


def ensure_default_policy(cidrs, subnets):
    policy_cidrs = netaddr.IPSet(cidrs)
    for subnet in subnets:
        subnet_cidr = netaddr.IPNetwork(subnet["cidr"])
        network_ip = subnet_cidr.network
        broadcast_ip = subnet_cidr.broadcast
        prefix_len = '32' if subnet_cidr.version == 4 else '128'
        default_policy_cidrs = ["%s/%s" % (network_ip, prefix_len),
                                "%s/%s" % (broadcast_ip, prefix_len)]
        for cidr in default_policy_cidrs:
            if (netaddr.IPNetwork(cidr) not in policy_cidrs
                    and cidr not in cidrs):
                cidrs.append(cidr)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db.quota.models import Quota
from neutron.db import quota_db


class QuarkQuotaDriver(quota_db.DbQuotaDriver):
    """Driver to perform necessary checks to enforce and obtain quotas.

    The default driver utilizes the local database.
    """

    @staticmethod
    def delete_tenant_quota(context, tenant_id):
        """Delete the quota entries for a given tenant_id.

        Atfer deletion, this tenant will use default quota values in conf.
        """

        tenant_quotas = context.session.query(Quota)
        tenant_quotas = tenant_quotas.filter_by(tenant_id=tenant_id)
        tenant_quotas.delete()

    @staticmethod
    def update_quota_limit(context, tenant_id, resource, limit):
        tenant_quota = context.session.query(Quota).filter_by(
            tenant_id=tenant_id, resource=resource).first()

        if tenant_quota:
            tenant_quota.update({'limit': limit})
        else:
            tenant_quota = Quota(tenant_id=tenant_id,
                                 resource=resource,
                                 limit=limit)
            context.session.add(tenant_quota)



import XenAPI

from quark.agent import xapi
from quark.tests import test_base

import mock


class TestVIF(test_base.TestBase):
    def test_str(self):
        rec = {"MAC": 2}
        self.assertEqual(str(xapi.VIF("1", rec, "3")), "1.2.3")

    def test_repr(self):
        rec = {"MAC": '2'}
        self.assertEqual(repr(xapi.VIF("1", rec, "3")), "VIF('1', '2', '3')")

    def test_eq(self):
        rec = {"MAC": '2'}
        self.assertEqual(xapi.VIF("1", rec, "3"), xapi.VIF("1", rec, "3"))

    def test_ne(self):
        rec1 = {"MAC": '2'}
        rec2 = {"MAC": '3'}
        vif1 = xapi.VIF("1", rec1, "4")
        vif2 = xapi.VIF("1", rec2, "4")
        vif3 = xapi.VIF("3", rec1, "4")
        vif4 = xapi.VIF("3", rec2, "4")

        self.assertNotEqual(vif1, vif2)
        self.assertNotEqual(vif1, vif3)
        self.assertNotEqual(vif1, vif4)


class TestXapiClient(test_base.TestBase):
    def setUp(self):
        patcher = mock.patch("quark.agent.xapi.XenAPI.Session")
        self.addCleanup(patcher.stop)
        self.session = patcher.start().return_value
        self.xclient = xapi.XapiClient()

    def test_get_instances(self):
        self.session.xenapi.VM.get_all_records.return_value = {
            "opaque1": {"other_config": {"nova_uuid": "uuid1"},
                        "power_state": "running", "is_a_template": False,
                        "is_control_domain": False,
                        "name_label": "instance-1",
                        "VIFs": ["opaque_vif1", "opaque_vif2"],
                        "domid": "1"},
            "opaque2": {"other_config": {},
                        "power_state": "", "is_a_template": True,
                        "is_control_domain": True,
                        "name_label": "",
                        "VIFs": [],
                        "domid": "2"},
        }
        instances = self.xclient.get_instances(self.session)
        vm = xapi.VM(ref="opaque1", uuid="uuid1",
                     dom_id="1", vifs=["opaque_vif1", "opaque_vif2"])
        self.assertEqual(instances, {"opaque1": vm})

    @mock.patch("quark.agent.xapi.XapiClient.get_instances")
    def test_get_interfaces(self, get_instances):
        instances = {"opaque1": xapi.VM(uuid="device_id1",
                                        ref="opaque1",
                                        vifs=["opaque_vif1"],
                                        dom_id="1")}
        get_instances.return_value = instances
        self.session.xenapi.VIF.get_all_records.return_value = {
            "opaque_vif1": {"VM": "opaque1", "MAC": "00:11:22:33:44:55"},
            "opaque_vif2": {"VM": "opaque2", "MAC": "55:44:33:22:11:00"},
        }
        interfaces = self.xclient.get_interfaces()
        rec = {"MAC": "00:11:22:33:44:55", "VM": "opaque1"}
        expected = set([xapi.VIF("device_id1", rec, "opaque_vif1")])
        self.assertEqual(interfaces, expected)

    @mock.patch("quark.agent.xapi.XapiClient.get_instances")
    def test_update_interfaces_added(self, get_instances):
        instances = {"opaque1": xapi.VM(uuid="device_id1",
                                        ref="opaque1",
                                        vifs=["opaque_vif1"],
                                        dom_id="1")}
        get_instances.return_value = instances
        rec = {"MAC": "00:11:22:33:44:55"}
        interfaces = [xapi.VIF("device_id1", rec,
                               "opaque_vif1")]

        dom_id = "1"
        vif_index = "0"

        vif_rec = {"device": vif_index, "VM": "opaqueref"}
        vm_rec = {"domid": dom_id}

        expected_args = dict(dom_id=dom_id, vif_index=vif_index)
        self.session.xenapi.VIF.get_record.return_value = vif_rec
        self.session.xenapi.VM.get_record.return_value = vm_rec
        self.xclient.update_interfaces(interfaces, [], [])

        xenapi_VIF = self.session.xenapi.VIF
        xenapi_VIF.add_to_other_config.assert_called_once_with(
            "opaque_vif1", "security_groups", "enabled")
        self.assertEqual(xenapi_VIF.remove_from_other_config.call_count, 0)

        self.session.xenapi.host.call_plugin.assert_called_once_with(
            self.session.xenapi.session.get_this_host.return_value,
            "neutron_vif_flow", "online_instance_flows",
            expected_args)

    def test_update_interfaces_added_vm_removed(self):
        rec = {"MAC": "00:11:22:33:44:55"}
        interfaces = [xapi.VIF("device_id1", rec,
                               "opaque_vif1")]
        vif_index = "0"

        vif_rec = {"device": vif_index, "VM": "opaqueref"}

        self.session.xenapi.VIF.get_record.return_value = vif_rec
        self.session.xenapi.VM.get_record.side_effect = XenAPI.Failure(
            "HANDLE_INVALID")

        self.xclient.update_interfaces(interfaces, [], [])

        xenapi_VIF = self.session.xenapi.VIF
        xenapi_VIF.add_to_other_config.assert_called_once_with(
            "opaque_vif1", "security_groups", "enabled")
        self.assertEqual(xenapi_VIF.remove_from_other_config.call_count, 0)

        self.assertEqual(self.session.xenapi.host.call_plugin.call_count, 0)

    def test_update_interfaces_updated(self):
        rec = {"MAC": "00:11:22:33:44:55"}
        interfaces = [xapi.VIF("device_id1", rec,
                               "opaque_vif1")]

        dom_id = "1"
        vif_index = "0"

        vif_rec = {"device": vif_index, "VM": "opaqueref"}
        vm_rec = {"domid": dom_id}

        expected_args = dict(dom_id=dom_id, vif_index=vif_index)
        self.session.xenapi.VIF.get_record.return_value = vif_rec
        self.session.xenapi.VM.get_record.return_value = vm_rec

        self.xclient.update_interfaces([], interfaces, [])

        xenapi_VIF = self.session.xenapi.VIF
        self.assertEqual(xenapi_VIF.add_to_other_config.call_count, 0)
        self.assertEqual(xenapi_VIF.remove_from_other_config.call_count, 0)

        expected_args = dict(dom_id="1", vif_index="0")
        self.session.xenapi.host.call_plugin.assert_called_once_with(
            self.session.xenapi.session.get_this_host.return_value,
            "neutron_vif_flow", "online_instance_flows",
            expected_args)

    def test_update_interfaces_removed(self):
        rec = {"MAC": "00:11:22:33:44:55"}
        interfaces = [xapi.VIF("device_id1", rec,
                               "opaque_vif1")]

        dom_id = "1"
        vif_index = "0"

        vif_rec = {"device": vif_index, "VM": "opaqueref"}
        vm_rec = {"domid": dom_id}

        expected_args = dict(dom_id=dom_id, vif_index=vif_index)
        self.session.xenapi.VIF.get_record.return_value = vif_rec
        self.session.xenapi.VM.get_record.return_value = vm_rec
        self.xclient.update_interfaces([], [], interfaces)

        xenapi_VIF = self.session.xenapi.VIF
        self.assertEqual(xenapi_VIF.add_to_other_config.call_count, 0)
        xenapi_VIF.remove_from_other_config.assert_called_once_with(
            "opaque_vif1", "security_groups")

        self.session.xenapi.host.call_plugin.assert_called_once_with(
            self.session.xenapi.session.get_this_host.return_value,
            "neutron_vif_flow", "online_instance_flows",
            expected_args)

    def test_update_interfaces_removed_vm_removed(self):
        rec = {"MAC": "00:11:22:33:44:55"}
        interfaces = [xapi.VIF("device_id1", rec,
                               "opaque_vif1")]

        self.session.xenapi.VIF.get_record.side_effect = XenAPI.Failure(
            "HANDLE_INVALID")
        self.session.xenapi.VM.get_record.return_value = XenAPI.Failure(
            "HANDLE_INVALID")
        self.xclient.update_interfaces([], [], interfaces)

        xenapi_VIF = self.session.xenapi.VIF
        self.assertEqual(xenapi_VIF.add_to_other_config.call_count, 0)
        xenapi_VIF.remove_from_other_config.assert_called_once_with(
            "opaque_vif1", "security_groups")

        self.assertEqual(self.session.xenapi.host.call_plugin.call_count, 0)

    def test_update_interfaces_none(self):
        self.xclient.update_interfaces([], [], [])

        xenapi_VIF = self.session.xenapi.VIF
        self.assertEqual(xenapi_VIF.remove_from_other_config.call_count, 0)
        self.assertEqual(xenapi_VIF.add_to_other_config.call_count, 0)

        self.assertEqual(self.session.xenapi.host.call_plugin.call_count, 0)

    def test_update_interfaces_removed_raises(self):
        rec = {"MAC": "00:11:22:33:44:55"}
        interfaces = [xapi.VIF("device_id1", rec,
                               "opaque_vif1")]

        dom_id = "1"
        vif_index = "0"

        vif_rec = {"device": vif_index, "VM": "opaqueref"}
        vm_rec = {"domid": dom_id}

        expected_args = dict(dom_id=dom_id, vif_index=vif_index)
        self.session.xenapi.VIF.get_record.return_value = vif_rec
        self.session.xenapi.VM.get_record.return_value = vm_rec
        xenapi_VIF = self.session.xenapi.VIF

        # Without the try/except in _unset, this raises and the test fails
        xenapi_VIF.remove_from_other_config.side_effect = XenAPI.Failure(
            "HANDLE_INVALID")

        self.xclient.update_interfaces([], [], interfaces)

        self.assertEqual(xenapi_VIF.add_to_other_config.call_count, 0)
        xenapi_VIF.remove_from_other_config.assert_called_once_with(
            "opaque_vif1", "security_groups")

        self.session.xenapi.host.call_plugin.assert_called_once_with(
            self.session.xenapi.session.get_this_host.return_value,
            "neutron_vif_flow", "online_instance_flows",
            expected_args)


class TestXapiSession(test_base.TestBase):
    def setUp(self):
        patcher = mock.patch("quark.agent.xapi.XenAPI.Session")
        self.addCleanup(patcher.stop)
        self.session = patcher.start().return_value

    @mock.patch("quark.agent.xapi.XapiClient._session")
    def test_sessioned_exception_handling(self, xapi_session):
        xapi_session.side_effect = XenAPI.Failure("HANDLE_INVALID")
        with self.assertRaises(XenAPI.Failure):
            xapi.XapiClient()
            self.session.logout.assert_called_once()



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for# the specific language governing permissions and limitations
#  under the License.

from neutron_lib import exceptions as n_exc

from quark.db import api as db_api
from quark import exceptions as q_exc
import quark.plugin_modules.segment_allocation_ranges as sa_ranges_api
from quark import segment_allocations
from quark.tests.functional.base import BaseFunctionalTest


class QuarkSegmentAllocationTest(BaseFunctionalTest):

    def setUp(self):
        super(QuarkSegmentAllocationTest, self).setUp()
        self.segment_type = 'vxlan'
        self.segment_id = 'segment_id'
        self.old_context = self.context
        self.context = self.context.elevated()

    def _make_segment_allocation_range_dict(self, segment_type=None,
                                            segment_id=None,
                                            first_id=1,
                                            last_id=5):
        if not segment_type:
            segment_type = self.segment_type

        if not segment_id:
            segment_id = self.segment_id

        return {
            'segment_type': segment_type,
            'segment_id': segment_id,
            'first_id': first_id,
            'last_id': last_id,
        }

    def _populate_segment_allocation_range(self, sa_range):
        """Populate a given segment range."""

        # Range of ids to allocate, first to last (inclusive)
        id_range = xrange(sa_range['first_id'],
                          sa_range['last_id'] + 1)

        sa_dicts = []
        total = 0
        for i in id_range:
            sa_dicts.append({
                'segment_id': sa_range['segment_id'],
                'segment_type': sa_range['segment_type'],
                'id': i,
                'segment_allocation_range_id': sa_range['id'],
                'deallocated': True
            })
            total = total + 1
        db_api.segment_allocation_range_populate_bulk(self.context, sa_dicts)
        self.context.session.flush()

        # assert our allocation were actually created
        allocs = db_api.segment_allocation_find(
            self.context, segment_allocation_range_id=sa_range['id']).all()
        self.assertEqual(len(allocs), len(id_range))

    def _create_segment_allocation_range(self, **kwargs):
        """Create a segment allocation range in the database."""
        sa_dict = self._make_segment_allocation_range_dict(**kwargs)
        sa_range = db_api.segment_allocation_range_create(
            self.context, **sa_dict)
        self.context.session.flush()
        self._populate_segment_allocation_range(sa_range)
        return sa_range

    def _allocate_segment(self, sa_range, count=1):
        """Populate a given segment range."""
        allocs = []
        for i in xrange(sa_range['first_id'], sa_range['first_id'] + count):
            filters = {
                'segment_allocation_range_id': sa_range['id'],
                'deallocated': True
            }
            alloc = db_api.segment_allocation_find(
                self.context, **filters).first()
            if not alloc:
                raise Exception("Could not find deallocated id.")
            update = {
                'deallocated': False
            }
            allocs.append(
                db_api.segment_allocation_update(
                    self.context, alloc, **update)
            )
            self.context.session.flush()
        self.assertEqual(len(allocs), count)
        return allocs

    def _sa_range_to_dict(self, sa_range, allocations=None):
        """Helper to turn a model into a dict for assertions."""
        size = (sa_range['last_id'] + 1) - sa_range['first_id']
        sa_range_dict = dict(sa_range)
        sa_range_dict.pop('created_at')
        sa_range_dict['size'] = size

        if allocations is not None:
            sa_range_dict['free_ids'] = size - allocations
        return sa_range_dict


class QuarkTestVXLANSegmentAllocation(QuarkSegmentAllocationTest):

    def setUp(self):
        super(QuarkTestVXLANSegmentAllocation, self).setUp()
        self.registry = segment_allocations.SegmentAllocationRegistry()
        self.driver = self.registry.get_strategy('vxlan')

    def test_segment_allocation(self):
        sa_range = self._create_segment_allocation_range()

        # assert we allocate and update correctly
        alloc = self.driver.allocate(
            self.context, sa_range['segment_id'], 'network_id_1')
        self.assertEqual(alloc['segment_type'], sa_range['segment_type'])
        self.assertEqual(alloc['segment_id'], sa_range['segment_id'])
        self.assertEqual(alloc['network_id'], 'network_id_1')

        # assert the remaining allocations remain unallocated
        allocs = db_api.segment_allocation_find(
            self.context).all()
        allocs.remove(alloc)
        self.assertEqual(len(allocs), 4)
        self.assertTrue(all([a["deallocated"] for a in allocs]))

        return sa_range, alloc

    def test_segment_deallocation(self):
        # We call the allocate test to set up an initial allocation
        # and assert that it actually worked.
        sa_range, alloc = self.test_segment_allocation()

        self.driver.deallocate(
            self.context, sa_range['segment_id'], 'network_id_1')

        # assert that our previous allocation is now free
        allocs = db_api.segment_allocation_find(
            self.context, id=alloc['id'],
            segment_id=sa_range['segment_id']).all()

        self.assertEqual(len(allocs), 1)
        self.assertTrue(allocs[0]["deallocated"])
        self.assertEqual(allocs[0]["network_id"], None)

    def test_allocation_segment_full(self):
        # create a range, and allocate everything
        sa_range = self._create_segment_allocation_range()
        self._allocate_segment(sa_range, count=5)

        self.assertRaises(
            q_exc.SegmentAllocationFailure,
            self.driver.allocate,
            self.context, sa_range['segment_id'], 'network_id_2')


class QuarkTestCreateSegmentAllocationRange(QuarkSegmentAllocationTest):

    def test_create_segment_allocation_range_unauthorized(self):
        sa_range_dict = self._make_segment_allocation_range_dict()
        sa_range_request = {"segment_allocation_range": sa_range_dict}

        self.assertRaises(
            n_exc.NotAuthorized,
            sa_ranges_api.create_segment_allocation_range,
            self.old_context, sa_range_request)

    def test_create_segment_allocation_range(self):
        """Assert a range is created."""

        # Create a segment allocation range
        sa_range_dict = self._make_segment_allocation_range_dict()
        sa_range_request = {"segment_allocation_range": sa_range_dict}
        sa_range = sa_ranges_api.create_segment_allocation_range(
            self.context, sa_range_request)

        # Find all ranges added in the db
        sa_range_models = db_api.segment_allocation_range_find(
            self.context, scope=db_api.ALL)
        # ensure non-admins can fetch them as well
        sa_range_models = db_api.segment_allocation_range_find(
            self.old_context, scope=db_api.ALL)

        # Assert we actually added the range to the db with correct
        # values and returned the correct response.
        self.assertEqual(len(sa_range_models), 1)
        self.assertEqual(self._sa_range_to_dict(sa_range_models[0]),
                         sa_range)

    def test_create_segment_allocation_range_invalid_fails(self):
        """Assert segments with invalid ranges are disallowed."""
        sa_range_dict = self._make_segment_allocation_range_dict()
        sa_range_request = {"segment_allocation_range": sa_range_dict}

        invalid_ranges = [
            (0, 5),  # first_id < MIN,
            (1, 2 ** 24 + 1),  # last_id > MAX,
            (5, 1),  # last_id < first_id,
            ('a', 5),  # invalid data
        ]
        for first_id, last_id in invalid_ranges:
            sa_range_dict['first_id'] = first_id
            sa_range_dict['last_id'] = last_id
            self.assertRaises(
                q_exc.InvalidSegmentAllocationRange,
                sa_ranges_api.create_segment_allocation_range,
                self.context, sa_range_request)

    def test_create_segment_allocation_range_creates_allocations(self):
        """Assert created segments populate the allocation table."""
        sa_range_dict = self._make_segment_allocation_range_dict()
        sa_range_request = {"segment_allocation_range": sa_range_dict}
        sa_range = sa_ranges_api.create_segment_allocation_range(
            self.context, sa_range_request)

        allocs = db_api.segment_allocation_find(
            self.context, segment_allocation_range_id=sa_range['id']).all()
        self.assertEqual(len(allocs), sa_range['size'])

    def test_create_segment_allocation_ranges(self):
        """Assert segments with same type/id are allowed."""
        sa_range_dict = self._make_segment_allocation_range_dict()
        sa_range_request = {"segment_allocation_range": sa_range_dict}

        valid_ranges = [
            (10, 15),
            (5, 9),
            (16, 20),
        ]
        for first_id, last_id in valid_ranges:
            sa_range_dict['first_id'] = first_id
            sa_range_dict['last_id'] = last_id
            sa_range = sa_ranges_api.create_segment_allocation_range(
                self.context, sa_range_request)

            # Find all ranges added in the db
            sa_range_models = db_api.segment_allocation_range_find(
                self.context, id=sa_range['id'], scope=db_api.ALL)

            # Assert we actually added the range to the db with correct
            # values and returned the correct response.
            self.assertEqual(len(sa_range_models), 1)
            self.assertEqual(self._sa_range_to_dict(sa_range_models[0]),
                             sa_range)

    def test_create_segment_allocation_ranges_diff_overlap_allowed(self):
        """Assert different segments with overlapping ids are allowed."""
        sa_range_dict = self._make_segment_allocation_range_dict()
        sa_range_request = {"segment_allocation_range": sa_range_dict}

        segment_ids = [
            'segment1',
            'segment2',
            'segment3'
        ]
        for segment_id in segment_ids:
            sa_range_dict['first_id'] = 1
            sa_range_dict['last_id'] = 5
            sa_range_dict['segment_id'] = segment_id
            sa_range = sa_ranges_api.create_segment_allocation_range(
                self.context, sa_range_request)

            # Find all ranges added in the db
            sa_range_models = db_api.segment_allocation_range_find(
                self.context, id=sa_range['id'], scope=db_api.ALL)

            # Assert we actually added the range to the db with correct
            # values and returned the correct response.
            self.assertEqual(len(sa_range_models), 1)
            self.assertEqual(self._sa_range_to_dict(sa_range_models[0]),
                             sa_range)

    def test_create_segment_allocation_ranges_same_overlap_fails(self):
        """Assert same segments with overlapping ids are disallowed."""
        sa_range_dict = self._make_segment_allocation_range_dict()
        sa_range_request = {"segment_allocation_range": sa_range_dict}

        # create initial segment with range 10-15
        sa_range_dict['first_id'] = 10
        sa_range_dict['last_id'] = 15
        sa_ranges_api.create_segment_allocation_range(
            self.context, sa_range_request)

        invalid_ranges = [
            (10, 15),  # same range
            (5, 10),  # collides at start
            (15, 20),  # collides at end
            (8, 12),  # overlaps from start
            (12, 17),  # overlaps from end
            (9, 16),  # superset
            (11, 14)  # subset
        ]
        for first_id, last_id in invalid_ranges:
            sa_range_dict['first_id'] = first_id
            sa_range_dict['last_id'] = last_id
            self.assertRaises(
                q_exc.InvalidSegmentAllocationRange,
                sa_ranges_api.create_segment_allocation_range,
                self.context, sa_range_request)


class QuarkTestGetSegmentAllocationRange(QuarkSegmentAllocationTest):

    def test_get_segment_allocation_range_unauthorized(self):
        sa_range = self._create_segment_allocation_range()

        self.assertRaises(
            n_exc.NotAuthorized,
            sa_ranges_api.get_segment_allocation_range,
            self.old_context, sa_range["id"])

    def test_get_segment_allocation_range_not_found(self):
        self._create_segment_allocation_range()

        self.assertRaises(
            n_exc.NotFound,
            sa_ranges_api.get_segment_allocation_range,
            self.context, "some_id")

    def test_get_segment_allocation_range(self):
        sa_range = self._create_segment_allocation_range()

        result = sa_ranges_api.get_segment_allocation_range(
            self.context, sa_range['id'])

        expected_result = self._sa_range_to_dict(sa_range, allocations=0)
        self.assertEqual(expected_result, result)

    def test_get_segment_allocation_range_with_allocations(self):
        sa_range = self._create_segment_allocation_range()
        allocs = self._allocate_segment(sa_range, count=2)

        result = sa_ranges_api.get_segment_allocation_range(
            self.context, sa_range['id'])

        expected_result = self._sa_range_to_dict(
            sa_range, allocations=len(allocs))
        self.assertEqual(expected_result, result)


class QuarkTestGetSegmentAllocationRanges(QuarkSegmentAllocationTest):

    def test_get_segment_allocation_ranges_unauthorized(self):
        self.assertRaises(
            n_exc.NotAuthorized,
            sa_ranges_api.get_segment_allocation_ranges,
            self.old_context)

    def test_get_segment_allocation_ranges_empty(self):
        result = sa_ranges_api.get_segment_allocation_ranges(self.context)
        self.assertEqual([], result)

    def test_get_segment_allocation_ranges(self):
        sa_range = self._create_segment_allocation_range()
        sa_range2 = self._create_segment_allocation_range(
            first_id=6, last_id=10)

        result = sa_ranges_api.get_segment_allocation_ranges(self.context)

        ex_result = [self._sa_range_to_dict(r) for r in [sa_range, sa_range2]]
        self.assertEqual(ex_result, result)


class QuarkTestDeleteSegmentAllocationRange(QuarkSegmentAllocationTest):

    def test_delete_segment_allocation_range_not_found(self):
        self._create_segment_allocation_range()

        self.assertRaises(
            n_exc.NotFound,
            sa_ranges_api.delete_segment_allocation_range,
            self.context, "some_id")

    def test_delete_segment_allocation_range_unauthorized(self):
        sa_range = self._create_segment_allocation_range()

        # assert non-admins are not authorized
        self.assertRaises(
            n_exc.NotAuthorized,
            sa_ranges_api.delete_segment_allocation_range,
            self.old_context, sa_range["id"])

        # assert the range was not deleted
        sa_ranges = db_api.segment_allocation_range_find(
            self.context, id=sa_range["id"], scope=db_api.ALL)
        self.assertEqual(sa_ranges, [sa_range])

    def test_delete_segment_allocation_range_in_use_fails(self):
        sa_range = self._create_segment_allocation_range()
        self._allocate_segment(sa_range, count=1)

        self.assertRaises(
            n_exc.InUse,
            sa_ranges_api.delete_segment_allocation_range,
            self.context, sa_range["id"])

        # assert the range was not deleted
        sa_ranges = db_api.segment_allocation_range_find(
            self.context, id=sa_range["id"], scope=db_api.ALL)
        self.assertEqual(sa_ranges, [sa_range])

    def test_delete_segment_allocation_range_deletes(self):
        sa_range = self._create_segment_allocation_range()
        sa_range_id = sa_range["id"]

        sa_ranges_api.delete_segment_allocation_range(
            self.context, sa_range_id)

        # assert that the range and it's unused allocations are deleted
        sa_range = db_api.segment_allocation_range_find(
            self.context, id=sa_range_id, scope=db_api.ALL)
        allocs = db_api.segment_allocation_find(
            self.context, segment_allocation_range_id=sa_range_id).all()
        self.assertEqual(sa_range, [])
        self.assertEqual(allocs, [])



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Optimized NVP client for Quark
"""

import aiclib
from oslo_log import log as logging

from quark.db import models
from quark.drivers.nvp_driver import NVPDriver

import sqlalchemy as sa
from sqlalchemy import orm

LOG = logging.getLogger(__name__)


class OptimizedNVPDriver(NVPDriver):
    def __init__(self):
        super(OptimizedNVPDriver, self).__init__()

    @classmethod
    def get_name(klass):
        return "NVP"

    def delete_network(self, context, network_id):
        lswitches = self._lswitches_for_network(context, network_id)
        for switch in lswitches:
            try:
                self._lswitch_delete(context, switch.nvp_id)
                self._remove_default_tz_bindings(
                    context, network_id)
            except aiclib.core.AICException as ae:
                LOG.info("LSwitch/Network %s found in database."
                         " Adding to orphaned database table."
                         % network_id)
                if ae.code != 404:
                    LOG.info("LSwitch/Network %s was found in NVP."
                             " Adding to orpaned table for later cleanup."
                             " Code: %s, Message: %s"
                             % (network_id, ae.code, ae.message))
                    orphaned_lswitch = OrphanedLSwitch(
                        nvp_id=switch.nvp_id,
                        network_id=switch.network_id,
                        display_name=switch.display_name
                    )
                    context.session.add(orphaned_lswitch)
                LOG.info("Deleting LSwitch/Network %s from original"
                         " table." % network_id)
                context.session.delete(switch)
            except Exception as e:
                message = e.args[0] if e.args else ''
                LOG.info("Failed to delete LSwitch/Network %s from "
                         " NVP (optimized). Message: %s"
                         % (network_id, message))

    def create_port(self, context, network_id, port_id,
                    status=True, security_groups=None,
                    device_id="", **kwargs):
        security_groups = security_groups or []
        nvp_port = super(OptimizedNVPDriver, self).create_port(
            context, network_id, port_id, status=status,
            security_groups=security_groups, device_id=device_id)
        switch_nvp_id = nvp_port["lswitch"]

        # slightly inefficient for the sake of brevity. Lets the
        # parent class do its thing then finds the switch that
        # the port was created on for creating the association. Switch should
        # be in the query cache so the subsequent lookup should be minimal,
        # but this could be an easy optimization later if we're looking.
        switch = self._lswitch_select_by_nvp_id(context, switch_nvp_id)

        new_port = LSwitchPort(port_id=nvp_port["uuid"],
                               switch_id=switch.id)
        context.session.add(new_port)
        switch.port_count = switch.port_count + 1
        return nvp_port

    def update_port(self, context, port_id, status=True,
                    security_groups=None, **kwargs):
        security_groups = security_groups or []
        mac_address = kwargs.get('mac_address')
        device_id = kwargs.get('device_id')
        nvp_port = super(OptimizedNVPDriver, self).update_port(
            context, port_id, mac_address=mac_address, device_id=device_id,
            status=status, security_groups=security_groups)
        port = self._lport_select_by_id(context, port_id)
        port.update(nvp_port)

    def delete_port(self, context, port_id, **kwargs):
        port = self._lport_select_by_id(context, port_id)
        if not port:
            LOG.warning("Lost local reference to NVP lport %s" % port_id)
            return  # we return here because there isn't anything else to do

        switch = port.switch
        try:
            self._lport_delete(context, port_id, switch)
        except aiclib.core.AICException as ae:
            LOG.info("LSwitchPort/Port %s found in database."
                     " Adding to orphaned database table."
                     % port_id)
            if ae.code != 404:
                LOG.info("LSwitchPort/Port %s was found in NVP."
                         " Adding to orpaned table for later cleanup."
                         " Code: %s, Message: %s"
                         % (port_id, ae.code, ae.args[0]))
                orphaned_lswitch_port = OrphanedLSwitchPort(
                    port_id=port_id,
                )
                context.session.add(orphaned_lswitch_port)
        except Exception as e:
            LOG.info("Failed to delete LSwitchPort/Port %s from "
                     " NVP (optimized). Message: %s"
                     % (port_id, e.args[0]))
        LOG.info("Deleting LSwitchPort/Port %s from original"
                 " table." % port_id)
        context.session.delete(port)
        switch.port_count = switch.port_count - 1
        if switch.port_count == 0:
            switches = self._lswitches_for_network(context, switch.network_id)
            if len(switches) > 1:  # do not delete last lswitch on network
                self._lswitch_delete(context, switch.nvp_id)

    def _lport_delete(self, context, port_id, switch=None):
        if switch is None:
            port = self._lport_select_by_id(context, port_id)
            switch = port.switch
        super(OptimizedNVPDriver, self).delete_port(
            context, port_id, lswitch_uuid=switch.nvp_id)

    def create_security_group(self, context, group_name, **group):
        nvp_group = super(OptimizedNVPDriver, self).create_security_group(
            context, group_name, **group)
        group_id = group.get('group_id')
        profile = SecurityProfile(id=group_id, nvp_id=nvp_group['uuid'])
        context.session.add(profile)

    def delete_security_group(self, context, group_id, **kwargs):
        super(OptimizedNVPDriver, self).delete_security_group(
            context, group_id)
        group = self._query_security_group(context, group_id)
        context.session.delete(group)

    def _lport_select_by_id(self, context, port_id):
        query = context.session.query(LSwitchPort)
        query = query.filter(LSwitchPort.port_id == port_id)
        return query.first()

    def _lswitch_delete(self, context, lswitch_uuid):
        switch = self._lswitch_select_by_nvp_id(context, lswitch_uuid)
        super(OptimizedNVPDriver, self)._lswitch_delete(
            context, lswitch_uuid)
        context.session.delete(switch)

    def _lswitch_select_by_nvp_id(self, context, nvp_id):
        switch = context.session.query(LSwitch).filter(
            LSwitch.nvp_id == nvp_id).first()
        return switch

    def _lswitch_select_first(self, context, network_id):
        query = context.session.query(LSwitch)
        query = query.filter(LSwitch.network_id == network_id)
        return query.first()

    def _lswitch_select_free(self, context, network_id):
        query = context.session.query(LSwitch)
        query = query.filter(LSwitch.port_count <
                             self.limits['max_ports_per_switch'])
        query = query.filter(LSwitch.network_id == network_id)
        switch = query.order_by(LSwitch.port_count).first()
        return switch

    def _lswitch_status_query(self, context, network_id):
        """Child implementation of lswitch_status_query.

        Deliberately empty as we rely on _get_network_details to be more
        efficient than we can be here.
        """
        pass

    def _lswitch_select_open(self, context, network_id=None, **kwargs):
        if self.limits['max_ports_per_switch'] == 0:
            switch = self._lswitch_select_first(context, network_id)
        else:
            switch = self._lswitch_select_free(context, network_id)
        if switch:
            return switch.nvp_id
        LOG.debug("Could not find optimized switch")

    def _get_network_details(self, context, network_id, switches):
        name, phys_net, phys_type, segment_id = None, None, None, None
        switch = self._lswitch_select_first(context, network_id)
        if switch:
            name = switch.display_name
            phys_net = switch.transport_zone
            phys_type = switch.transport_connector
            segment_id = switch.segment_id
            return dict(network_name=name, phys_net=phys_net,
                        phys_type=phys_type, segment_id=segment_id)

    def _lswitch_create(self, context, network_name=None, tags=None,
                        network_id=None, **kwargs):
        nvp_id = super(OptimizedNVPDriver, self)._lswitch_create(
            context, network_name, tags, network_id, **kwargs)
        return self._lswitch_create_optimized(context, network_name, nvp_id,
                                              network_id, **kwargs).nvp_id

    def _lswitch_create_optimized(self, context, network_name, nvp_id,
                                  network_id, phys_net=None, phys_type=None,
                                  segment_id=None):
        new_switch = LSwitch(nvp_id=nvp_id, network_id=network_id,
                             port_count=0, transport_zone=phys_net,
                             transport_connector=phys_type,
                             display_name=network_name[:40],
                             segment_id=segment_id)
        context.session.add(new_switch)
        return new_switch

    def get_lswitch_ids_for_network(self, context, network_id):
        """Public interface for fetching lswitch ids for a given network.

        NOTE(morgabra) This is here because calling private methods
        from outside the class feels wrong, and we need to be able to
        fetch lswitch ids for use in other drivers.
        """
        lswitches = self._lswitches_for_network(context, network_id)
        return [s['nvp_id'] for s in lswitches]

    def _lswitches_for_network(self, context, network_id):
        switches = context.session.query(LSwitch).filter(
            LSwitch.network_id == network_id).all()
        return switches

    def _lswitch_from_port(self, context, port_id):
        port = self._lport_select_by_id(context, port_id)
        return port.switch.nvp_id

    def _query_security_group(self, context, group_id):
        return context.session.query(SecurityProfile).filter(
            SecurityProfile.id == group_id).first()

    def _make_security_rule_dict(self, rule):
        res = {"port_range_min": rule.get("port_range_min"),
               "port_range_max": rule.get("port_range_max"),
               "protocol": rule.get("protocol"),
               "ip_prefix": rule.get("remote_ip_prefix"),
               "group_id": rule.get("remote_group_id"),
               "ethertype": rule.get("ethertype")}
        for key, value in res.items():
            if value is None:
                res.pop(key)
        return res

    def _get_security_group(self, context, group_id):
        group = context.session.query(models.SecurityGroup).filter(
            models.SecurityGroup.id == group_id).first()
        rulelist = {'ingress': [], 'egress': []}
        for rule in group.rules:
            rulelist[rule.direction].append(
                self._make_security_rule_dict(rule))
        return {'uuid': self._query_security_group(context, group_id).nvp_id,
                'logical_port_ingress_rules': rulelist['ingress'],
                'logical_port_egress_rules': rulelist['egress']}

    def _check_rule_count_per_port(self, context, group_id):
        ports = context.session.query(models.SecurityGroup).filter(
            models.SecurityGroup.id == group_id).first().get('ports', [])
        groups = (set(group.id for group in port.get('security_groups', []))
                  for port in ports)
        return max(self._check_rule_count_for_groups(
            context, (self._get_security_group(context, id) for id in g))
            for g in groups)


class LSwitchPort(models.BASEV2, models.HasId):
    __tablename__ = "quark_nvp_driver_lswitchport"
    port_id = sa.Column(sa.String(36), nullable=False, index=True)
    switch_id = sa.Column(sa.String(36),
                          sa.ForeignKey("quark_nvp_driver_lswitch.id"),
                          nullable=False)


class LSwitch(models.BASEV2, models.HasId):
    __tablename__ = "quark_nvp_driver_lswitch"
    nvp_id = sa.Column(sa.String(36), nullable=False, index=True)
    network_id = sa.Column(sa.String(36), nullable=False, index=True)
    display_name = sa.Column(sa.String(255))
    port_count = sa.Column(sa.Integer())
    ports = orm.relationship(LSwitchPort, backref='switch')
    transport_zone = sa.Column(sa.String(36))
    transport_connector = sa.Column(sa.String(20))
    segment_id = sa.Column(sa.Integer())


class QOS(models.BASEV2, models.HasId):
    __tablename__ = "quark_nvp_driver_qos"
    display_name = sa.Column(sa.String(255), nullable=False)
    max_bandwidth_rate = sa.Column(sa.Integer(), nullable=False)
    min_bandwidth_rate = sa.Column(sa.Integer(), nullable=False)


class SecurityProfile(models.BASEV2, models.HasId):
    __tablename__ = "quark_nvp_driver_security_profile"
    nvp_id = sa.Column(sa.String(36), nullable=False, index=True)


class OrphanedLSwitch(models.BASEV2, models.HasId):
    __tablename__ = "quark_nvp_orphaned_lswitches"
    nvp_id = sa.Column(sa.String(36), nullable=False, index=True)
    network_id = sa.Column(sa.String(36), nullable=False, index=True)
    display_name = sa.Column(sa.String(255), index=True)


class OrphanedLSwitchPort(models.BASEV2, models.HasId):
    __tablename__ = "quark_nvp_orphaned_lswitch_ports"
    port_id = sa.Column(sa.String(36), nullable=False, index=True)



import netaddr

from quark.db import api as db_api
from quark.tests.functional.base import BaseFunctionalTest


class QuarkIPPoliciesSizeTest(BaseFunctionalTest):
    def test_ip_policies_create(self):
        ip_policy_dict = dict(
            exclude=["192.168.10.0/32", "192.168.10.255/32"])
        ip_policy = db_api.ip_policy_create(self.context, **ip_policy_dict)
        self.assertEqual(ip_policy["size"], 2)

    def test_ip_policies_update(self):
        ip_policy_dict = dict(
            exclude=["192.168.10.0/32", "192.168.10.255/32"])
        ip_policy = db_api.ip_policy_create(self.context, **ip_policy_dict)
        ip_policy_update_dict = dict(
            exclude=["192.168.10.0/32", "192.168.10.13/32",
                     "192.168.10.255/32"])
        updated_ip_policy = db_api.ip_policy_update(
            self.context, ip_policy, **ip_policy_update_dict)
        self.assertEqual(updated_ip_policy["size"], 3)


class QuarkIPPoliciesFirstLastIPTest(BaseFunctionalTest):
    def test_ip_policies_create(self):
        exclude_first_last = dict()
        for cidr in ("192.168.10.0/32", "192.168.10.255/32"):
            exclude_first_last[cidr] = netaddr.IPNetwork(cidr).ipv6().first
        ip_policy = db_api.ip_policy_create(
            self.context, exclude=exclude_first_last)
        self.assertEqual(len(ip_policy["exclude"]), 2)
        for ippc in ip_policy["exclude"]:
            self.assertEqual(exclude_first_last[ippc["cidr"]],
                             ippc["first_ip"])
            self.assertEqual(exclude_first_last[ippc["cidr"]],
                             ippc["last_ip"])

    def test_ip_policies_update(self):
        ip_policy_dict = dict(
            exclude=["192.168.10.0/32", "192.168.10.255/32"])
        ip_policy = db_api.ip_policy_create(self.context, **ip_policy_dict)
        new_exclude_first_last = dict()
        for cidr in ("192.168.10.0/32", "192.168.10.13/32",
                     "192.168.10.255/32"):
            new_exclude_first_last[cidr] = netaddr.IPNetwork(cidr).ipv6().first
        updated_ip_policy = db_api.ip_policy_update(
            self.context, ip_policy, exclude=new_exclude_first_last.keys())
        self.assertEqual(len(updated_ip_policy["exclude"]), 3)
        for ippc in updated_ip_policy["exclude"]:
            self.assertEqual(new_exclude_first_last[ippc["cidr"]],
                             ippc["first_ip"])
            self.assertEqual(new_exclude_first_last[ippc["cidr"]],
                             ippc["last_ip"])



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib

import mock
import netaddr
from neutron_lib import exceptions as n_exc

from quark.db import api as db_api
from quark import exceptions as q_exc
from quark.tests import test_quark_plugin


class TestQuarkGetRoutes(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, routes):
        with mock.patch("quark.db.api.route_find") as route_find:
            route_find.return_value = routes
            yield

    def test_get_routes(self):
        route = dict(id=1, cidr="192.168.0.0/24", gateway="192.168.0.1",
                     subnet_id=2)
        with self._stubs(routes=[route]):
            res = self.plugin.get_routes(self.context)
            for key in route.keys():
                self.assertEqual(res[0][key], route[key])

    def test_get_route(self):
        route = dict(id=1, cidr="192.168.0.0/24", gateway="192.168.0.1",
                     subnet_id=2)
        with self._stubs(routes=route):
            res = self.plugin.get_route(self.context, 1)
            for key in route.keys():
                self.assertEqual(res[key], route[key])

    def test_get_route_not_found_fails(self):
        with self._stubs(routes=None):
            with self.assertRaises(q_exc.RouteNotFound):
                self.plugin.get_route(self.context, 1)


class TestQuarkCreateRoutes(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, create_route, find_routes, subnet):
        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.route_create" % db_mod),
            mock.patch("%s.route_find" % db_mod),
            mock.patch("%s.subnet_find" % db_mod),
            # This module can't run independently otherwise, as the Quota
            # model isn't defined in this test unit, and no clean way to
            # use a model such that it would be defined. In other words,
            # running the other tests still has side-effects, which should
            # be investigated and cleaned up later.
            mock.patch("neutron.quota.QuotaEngine.limit_check")
        ) as (route_create, route_find, subnet_find, quota):
            route_create.return_value = create_route
            route_find.return_value = find_routes
            subnet_find.return_value = subnet
            yield

    def test_create_route_no_cidr_raises(self):
        subnet = dict(id=2, ip_policy=[], cidr="192.168.0.0/24")
        create_route = dict(id=1, gateway="192.168.0.1",
                            subnet_id=subnet["id"])
        with self._stubs(create_route=create_route, find_routes=[],
                         subnet=subnet):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_route(self.context,
                                         dict(route=create_route))

    def test_create_route_no_gateway_raises(self):
        subnet = dict(id=2, ip_policy=[], cidr="192.168.0.0/24")
        create_route = dict(id=1, cidr="192.168.0.0/24",
                            subnet_id=subnet["id"])
        with self._stubs(create_route=create_route, find_routes=[],
                         subnet=subnet):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_route(self.context,
                                         dict(route=create_route))

    def test_create_route_no_subnet_fails(self):
        excluded_net = netaddr.IPNetwork("192.168.0.1/32")
        ip_policy = {"exclude": [excluded_net]}
        subnet = dict(id=2, ip_policy=ip_policy, cidr="192.168.0.0/24")
        route = dict(id=1, cidr="192.168.0.0/24", gateway="192.168.0.1",
                     subnet_id=subnet["id"])
        with self._stubs(create_route=route, find_routes=[], subnet=None):
            with self.assertRaises(n_exc.SubnetNotFound):
                self.plugin.create_route(self.context, dict(route=route))

    def test_create_route_no_subnet_raises(self):
        subnet = dict(id=2, ip_policy=[], cidr="192.168.0.0/24")
        create_route = dict(id=1, cidr="192.168.0.0/24", gateway="192.168.0.1")
        with self._stubs(create_route=create_route, find_routes=[],
                         subnet=subnet):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_route(self.context,
                                         dict(route=create_route))


class TestQuarkDeleteRoutes(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, route):
        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.route_delete" % db_mod),
            mock.patch("%s.route_find" % db_mod),
        ) as (route_delete, route_find):
            route_find.return_value = route
            yield route_delete, route_find

    def test_delete_route(self):
        route = dict(id=1, cidr="192.168.0.0/24", gateway="192.168.0.1",
                     subnet_id=2)
        with self._stubs(route=route) as (route_delete, route_find):
            self.plugin.delete_route(self.context, 1)
            self.assertTrue(route_delete.called)

    def test_delete_route_not_found_fails(self):
        with self._stubs(route=None):
            with self.assertRaises(q_exc.RouteNotFound):
                self.plugin.delete_route(self.context, 1)

    def test_delete_route_deletes_correct_route(self):
        route = dict(id=1, cidr="192.168.0.0/24", gateway="192.168.0.1",
                     subnet_id=2)
        with self._stubs(route=route) as (route_delete, route_find):
            self.plugin.delete_route(self.context, 1)
            self.assertTrue(route_find.called_with(context=self.context, id=1,
                                                   scope=db_api.ONE))
            self.assertTrue(route_delete.called)



import os

from neutron import context
from neutron.db import api as neutron_db_api
from oslo_config import cfg
from sqlalchemy.orm import configure_mappers

from quark.db import models
from quark import quota_driver
from quark.tests import test_base


class MySqlBaseFunctionalTest(test_base.TestBase):
    @classmethod
    def setUpClass(cls):
        default = 'mysql://root@localhost/quark_functional_tests'
        sql_string = os.getenv('QUARK_MYSQL_TESTS_URL', default)
        cfg.CONF.set_override(
            'connection',
            sql_string,
            'database')
        cfg.CONF.set_override(
            'max_retries',
            1,
            'database')
        cfg.CONF.set_override(
            'retry_interval',
            0,
            'database')
        cfg.CONF.set_override(
            'connection_debug',
            '0',
            'database')

    def setUp(self):
        super(MySqlBaseFunctionalTest, self).setUp()
        self.context = context.Context('fake', 'fake', is_admin=False)
        configure_mappers()
        engine = neutron_db_api.get_engine()
        models.BASEV2.metadata.create_all(engine)
        quota_driver.Quota.metadata.create_all(engine)

    def tearDown(self):
        engine = neutron_db_api.get_engine()
        models.BASEV2.metadata.drop_all(engine)
        quota_driver.Quota.metadata.drop_all(engine)



# Copyright (c) 2013 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.api import extensions
from neutron.api.v2 import attributes as attr
from neutron.api.v2 import resource_helper
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


def _validate_list_of_port_dicts(values, data):
    if not isinstance(values, list):
        msg = _("'%s' is not a list") % data
        return msg

    for item in values:
        msg = _validate_port_dict(item)
        if msg:
            return msg

    items = [tuple(entry.items()) for entry in values]
    if len(items) != len(set(items)):
        msg = _("Duplicate items in the list: '%s'") % values
        return msg


def _validate_port_dict(values):
    if not isinstance(values, dict):
        msg = _("%s is not a valid dictionary") % values
        LOG.debug(msg)
        return msg
    port_id = values.get('port_id')
    fixed_ip = values.get('fixed_ip_address')
    msg = attr._validate_uuid(port_id)
    if msg:
        return msg
    if fixed_ip is None:
        return
    msg = attr._validate_ip_address(fixed_ip)
    if msg:
        return msg

attr.validators['type:validate_list_of_port_dicts'] = (
    _validate_list_of_port_dicts
)

RESOURCE_NAME = "scalingip"
RESOURCE_COLLECTION = RESOURCE_NAME + "s"

RESOURCE_ATTRIBUTE_MAP = {
    RESOURCE_COLLECTION: {
        'id': {
            'allow_post': False, 'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'primary_key': True
        },
        "scaling_ip_address": {
            'allow_post': True, 'allow_put': False,
            'validate': {'type:ip_address_or_none': None},
            'is_visible': True, 'default': None,
            'enforce_policy': True
        },
        "tenant_id": {
            'allow_post': True, 'allow_put': False,
            'required_by_policy': True,
            'validate': {'type:string': attr.TENANT_ID_MAX_LEN},
            'is_visible': True
        },
        "scaling_network_id": {
            'allow_post': True, 'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True
        },
        "ports": {
            'allow_post': True, 'allow_put': True,
            'validate': {
                'type:validate_list_of_port_dicts': None
            },
            'is_visible': True,
            'required_by_policy': True
        }
    }
}


class Scalingip(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return RESOURCE_NAME

    @classmethod
    def get_alias(cls):
        return RESOURCE_NAME

    @classmethod
    def get_description(cls):
        return "Scaling IPs"

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/network/ext/"
                "networks_quark/api/v2.0")

    @classmethod
    def get_updated(cls):
        return "2016-01-20T19:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        attr.PLURALS.update(plural_mappings)
        return resource_helper.build_resource_info(plural_mappings,
                                                   RESOURCE_ATTRIBUTE_MAP,
                                                   None,
                                                   register_quota=True)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import mock
import netaddr
from neutron.common import rpc
from neutron_lib import exceptions as n_exc

from quark.db import api as db_api
from quark.db import models
import quark.ipam

# import below necessary if file run by itself
from quark import plugin  # noqa
import quark.plugin_modules.ip_policies as policy_api
import quark.plugin_modules.networks as network_api
import quark.plugin_modules.subnets as subnet_api
from quark.tests.functional.base import BaseFunctionalTest

from oslo_config import cfg


class QuarkIpamBaseFunctionalTest(BaseFunctionalTest):
    def setUp(self):
        super(QuarkIpamBaseFunctionalTest, self).setUp()

        patcher = mock.patch("neutron.common.rpc.oslo_messaging")
        patcher.start()
        self.addCleanup(patcher.stop)
        rpc.init(mock.MagicMock())


class QuarkIPAddressAllocate(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnet):
        self.ipam = quark.ipam.QuarkIpamANY()
        with self.context.session.begin():
            next_ip = subnet.pop("next_auto_assign_ip", 0)
            net_mod = db_api.network_create(self.context, **network)
            subnet["network"] = net_mod
            sub_mod = db_api.subnet_create(self.context, **subnet)
            # NOTE(asadoughi): update after cidr constructor has been invoked
            db_api.subnet_update(self.context,
                                 sub_mod,
                                 next_auto_assign_ip=next_ip)
        yield net_mod
        with self.context.session.begin():
            db_api.subnet_delete(self.context, sub_mod)
            db_api.network_delete(self.context, net_mod)

    def test_allocate_finds_no_deallocated_creates_new_ip(self):
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 2
        subnet = dict(id=1, cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                      ip_policy=None, tenant_id="fake")
        with self._stubs(network, subnet) as net:
            ipaddress = []
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net["id"], 0, 0)
            self.assertIsNotNone(ipaddress[0]['id'])
            self.assertEqual(ipaddress[0]['address'], 281470681743362)
            self.assertEqual(ipaddress[0]['version'], 4)
            self.assertEqual(ipaddress[0]['used_by_tenant_id'], "fake")


class TestQuarkIpamAllocateFromV6Subnet(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnet, ip_policy):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(mock.patch("neutron.common.rpc.get_notifier")):
            net = network_api.create_network(self.context, network)
            subnet['subnet']['network_id'] = net['id']
            sub = subnet_api.create_subnet(self.context, subnet)
            ipp = policy_api.update_ip_policy(self.context,
                                              sub["ip_policy_id"], ip_policy)
            sub = subnet_api.get_subnet(self.context, sub['id'])
        yield net, sub, ipp

    def test_allocate_v6_with_mac_fails_policy_raises(self):
        cidr = netaddr.IPNetwork("fe80::dead:beef/64")
        allocation_pool = [{"start": cidr[-4], "end": cidr[-2]}]
        subnet = dict(allocation_pools=allocation_pool,
                      cidr="fe80::dead:beef/64", ip_version=6,
                      next_auto_assign_ip=0, tenant_id="fake")
        subnet = {"subnet": subnet}

        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}

        ip_policy = {"exclude": ["fe80::dead:beef/64"]}
        ip_policy = {"ip_policy": ip_policy}

        mac = models.MacAddress()
        mac["address"] = netaddr.EUI("AA:BB:CC:DD:EE:FF")

        old_override = cfg.CONF.QUARK.v6_allocation_attempts
        cfg.CONF.set_override('v6_allocation_attempts', 1, 'QUARK')

        with self._stubs(network, subnet, ip_policy) as (net, sub, ipp):
            with self.assertRaises(n_exc.IpAddressGenerationFailure):
                self.ipam.allocate_ip_address(self.context, [], net["id"], 0,
                                              0, subnets=[sub["id"]])
        cfg.CONF.set_override('v6_allocation_attempts', old_override, 'QUARK')


class QuarkIPAddressFindReallocatable(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnet):
        self.ipam = quark.ipam.QuarkIpamANY()
        with self.context.session.begin():
            next_ip = subnet.pop("next_auto_assign_ip", 0)
            net_mod = db_api.network_create(self.context, **network)
            subnet["network"] = net_mod
            sub_mod = db_api.subnet_create(self.context, **subnet)
            # NOTE(asadoughi): update after cidr constructor has been invoked
            db_api.subnet_update(self.context,
                                 sub_mod,
                                 next_auto_assign_ip=next_ip)
        yield net_mod

    def test_find_reallocatable_ips_does_not_raise(self):
        """Regression testing

        A patch recently introduced a bug wherein addressses
        could not be returned to the ip_address_find call in
        attempt_to_reallocate_ip. Adding this test to prevent
        a future regression.
        """

        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 2
        subnet = dict(id=1, cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                      ip_policy=None, tenant_id="fake")

        with self._stubs(network, subnet) as net:
            ip_kwargs = {
                "network_id": net["id"], "reuse_after": 14400,
                "deallocated": True, "scope": db_api.ONE,
                "lock_mode": True, "version": 4,
                "order_by": "address"}

            try:
                db_api.ip_address_find(self.context, **ip_kwargs)
            except Exception:
                self.fail("This should not have raised")


class QuarkIPAddressAllocateWithFullSubnetsNotMarkedAsFull(
        QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _fixtures(self, models):

        self.ipam = quark.ipam.QuarkIpamANY()
        net = dict(name="public", tenant_id='fake')
        net_mod = db_api.network_create(self.context, **net)
        with self.context.session.begin():
            for model in models:
                policy_mod = db_api.ip_policy_create(
                    self.context, **model['ip_policy'])
                model['subnet']["network"] = net_mod
                model['subnet']["ip_policy"] = policy_mod
                next_ip = model['subnet'].pop("next_auto_assign_ip", 0)
                sub_mod = db_api.subnet_create(self.context, **model['subnet'])
                # NOTE(amir): update after cidr constructor has been invoked
                db_api.subnet_update(self.context,
                                     sub_mod,
                                     next_auto_assign_ip=next_ip)
        yield net_mod

    def _create_models(self, subnet_cidr, ip_version, next_ip):
        models = {}
        net = netaddr.IPNetwork(subnet_cidr)
        first = str(netaddr.IPAddress(net.first)) + "/32"
        last = str(netaddr.IPAddress(net.last)) + "/32"
        models['ip_policy'] = dict(name='testpolicy',
                                   description='blah',
                                   exclude=[first, last])
        models["subnet"] = dict(cidr=subnet_cidr,
                                next_auto_assign_ip=next_ip,
                                tenant_id='fake',
                                do_not_use=False)
        return models

    def test_subnets_get_marked_as_full_retroactively(self):
        models = []
        models.append(self._create_models(
            "0.0.0.0/31",
            4,
            netaddr.IPNetwork("0.0.0.0/31").ipv6().last))
        models.append(self._create_models(
            "1.1.1.0/31",
            4,
            netaddr.IPNetwork("1.1.1.0/31").ipv6().last))
        models.append(self._create_models(
            "2.2.2.0/30",
            4,
            netaddr.IPNetwork("2.2.2.0/30").ipv6().first))

        with self._fixtures(models) as net:
            ipaddress = []
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net['id'], 0, 0)
            self.assertEqual(ipaddress[0].version, 4)
            self.assertEqual(ipaddress[0].address_readable, "2.2.2.1")

            with self.context.session.begin():
                subnets = db_api.subnet_find(self.context, None, None, None,
                                             False).all()
                self.assertEqual(len(subnets), 3)

                full_subnets = [s for s in subnets
                                if s.next_auto_assign_ip == -1]
                self.assertEqual(len(full_subnets), 2)
                available_subnets = list(set(full_subnets) ^ set(subnets))
                self.assertEqual(len(available_subnets), 1)
                self.assertEqual(available_subnets[0].cidr, "2.2.2.0/30")
                self.assertEqual(available_subnets[0].next_auto_assign_ip,
                                 netaddr.IPAddress("2.2.2.2").ipv6().value)


class QuarkIPAddressReallocateAllocated(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnets, ipam_strategy):
        self.ipam = ipam_strategy
        with self.context.session.begin():
            net_mod = db_api.network_create(self.context, **network)
            next_ip = []
            sub_mod = []
            for sub in subnets:
                next_ip.append(sub.pop("next_auto_assign_ip", 0))
                sub["network"] = net_mod
                sub_mod.append(db_api.subnet_create(self.context, **sub))
            for sub, ip_next in zip(sub_mod, next_ip):
                # NOTE(asadoughi): update after cidr constructor has been
                # invoked
                db_api.subnet_update(self.context,
                                     sub,
                                     next_auto_assign_ip=ip_next)
        yield net_mod, sub_mod
        with self.context.session.begin():
            for sub in sub_mod:
                db_api.subnet_delete(self.context, sub)
            db_api.network_delete(self.context, net_mod)

    def test_allocate_ipv4_throws_cannot_reallocate_allocated_ip(self):
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 2
        subnet = dict(id=1, cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                      ip_policy=None, tenant_id="fake")
        subnets = [subnet]
        ipam_strategy = quark.ipam.QuarkIpamANY()
        with self._stubs(network, subnets, ipam_strategy) as (net, sub):
            ipaddress = []
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net["id"], 0, 0)
            self.assertIsNotNone(ipaddress[0]['id'])
            self.assertEqual(ipaddress[0]['address'], 281470681743362)
            self.assertEqual(ipaddress[0]['version'], 4)
            self.assertEqual(ipaddress[0]['used_by_tenant_id'], "fake")
            # Attempt to allocate the same IP
            with self.assertRaises(n_exc.IpAddressInUse):
                allocated_ip = [ipaddress[0]['address']]
                self.ipam.allocate_ip_address(self.context, [],
                                              net["id"], 0, 0,
                                              ip_addresses=allocated_ip)

    def test_allocate_ipv6_throws_cannot_reallocate_allocated_ip(self):
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 2
        subnet = dict(id=1, cidr="fe80::dead:beef/64",
                      next_auto_assign_ip=next_ip,
                      ip_policy=None, tenant_id="fake")
        subnets = [subnet]
        ipam_strategy = quark.ipam.QuarkIpamANY()
        with self._stubs(network, subnets, ipam_strategy) as (net, sub):
            ipaddress = []
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net["id"], 0, 0, version=6)
            self.assertIsNotNone(ipaddress[0]['id'])
            self.assertEqual(ipaddress[0]['version'], 6)
            self.assertEqual(ipaddress[0]['used_by_tenant_id'], "fake")
            # Attempt to allocate the same IP
            with self.assertRaises(n_exc.IpAddressInUse):
                allocated_ip = [ipaddress[0]['address']]
                self.ipam.allocate_ip_address(self.context, [],
                                              net["id"], 0, 0,
                                              ip_addresses=allocated_ip)

    def test_allocate_both_v4_v6_throws_cannot_reallocate_allocated_ip(self):
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 2
        subnet1 = dict(id=1, cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                       ip_policy=None, tenant_id="fake", version=4)
        subnet2 = dict(id=2, cidr="fe80::dead:beef/64",
                       next_auto_assign_ip=next_ip,
                       ip_policy=None, tenant_id="fake", version=6)
        subnets = [subnet1, subnet2]
        ipam_strategy = quark.ipam.QuarkIpamBOTHREQ()
        with self._stubs(network, subnets, ipam_strategy) as (net, sub):
            ipaddress = []
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net["id"], 0, 0, subnets=[1, 2])
            self.assertEqual(len(ipaddress), 2)
            for ip in ipaddress:
                self.assertTrue(ip['version'] in [4, 6])
                self.assertIsNotNone(ip['id'])
                self.assertEqual(ip['used_by_tenant_id'], 'fake')
            # Attempt to allocate the same IP
            with self.assertRaises(n_exc.IpAddressInUse):
                allocated_ip = [ip['address_readable'] for ip in ipaddress]
                self.ipam.allocate_ip_address(self.context, [],
                                              net["id"], 0, 0,
                                              ip_addresses=allocated_ip,
                                              subnets=[1, 2])


class QuarkIPAddressReallocateDeallocated(QuarkIpamBaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnets, ipam_strategy):
        self.ipam = ipam_strategy
        with self.context.session.begin():
            net_mod = db_api.network_create(self.context, **network)
            next_ip = []
            sub_mod = []
            for sub in subnets:
                next_ip.append(sub.pop("next_auto_assign_ip", 0))
                sub["network"] = net_mod
                sub_mod.append(db_api.subnet_create(self.context, **sub))
            for sub, ip_next in zip(sub_mod, next_ip):
                # NOTE(asadoughi): update after cidr constructor has been
                # invoked
                db_api.subnet_update(self.context,
                                     sub,
                                     next_auto_assign_ip=ip_next)
        yield net_mod, sub_mod
        with self.context.session.begin():
            for sub in sub_mod:
                db_api.subnet_delete(self.context, sub)
            db_api.network_delete(self.context, net_mod)

    def test_allocate_deallocated_ips_ipam_both_req(self):
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 2
        subnet1 = dict(id=1, cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                       ip_policy=None, tenant_id="fake", version=4)
        subnet2 = dict(id=2, cidr="fe80::dead:beef/64",
                       next_auto_assign_ip=next_ip,
                       ip_policy=None, tenant_id="fake", version=6)
        subnets = [subnet1, subnet2]
        ipam_strategy = quark.ipam.QuarkIpamBOTHREQ()
        with self._stubs(network, subnets, ipam_strategy) as (net, sub):
            ipaddress = []
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net["id"], 0, 0, subnets=[1, 2])
            self.assertEqual(len(ipaddress), 2)
            for ip in ipaddress:
                self.assertTrue(ip['version'] in [4, 6])
                self.assertIsNotNone(ip['id'])
                self.assertEqual(ip['used_by_tenant_id'], 'fake')
            # Deallocate both given ip's
            for ip in ipaddress:
                self.ipam.deallocate_ip_address(self.context, ip)

            # Now attempt to reallocate
            self.ipam.allocate_ip_address(self.context, ipaddress,
                                          net["id"], 0, 0, subnets=[1, 2])



# Copyright (c) 2013 Rackspace Hosting Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions

EXTENDED_ATTRIBUTES_2_0 = {
    "subnets": {
        "segment_id": {"allow_post": True, "allow_put": False,
                       "default": None},
        "ip_policy_id": {"allow_post": False, "allow_put": False,
                         "default": None, "is_visible": True}
    }
}


class Subnets_quark(extensions.ExtensionDescriptor):
    """Extends subnets for quark API purposes.

    * Shunts enable_dhcp to false
    """

    @classmethod
    def get_name(cls):
        return "Quark Subnets API Extension"

    @classmethod
    def get_alias(cls):
        return "subnets_quark"

    @classmethod
    def get_description(cls):
        return "Quark Subnets API Extension"

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/api/openstack-network/2.0/content/"
                "Subnets.html")

    @classmethod
    def get_updated(cls):
        return "2013-04-22T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}



#!/usr/bin/env python
# Copyright 2016 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import sys

from gunicorn.app import base
from gunicorn import config as gconfig
from neutron.common import config
from neutron.common import utils
from neutron import service  # noqa  For api_workers config value
from oslo_config import cfg
from oslo_log import log as logging


options = [
    cfg.StrOpt('access_log',
               default='/var/log/neutron/http_access.log',
               help='The Access log file to write to.'),
    cfg.StrOpt('error_log',
               default='/var/log/neutron/http_error.log',
               help='The Error log file to write to.'),
    cfg.StrOpt('worker_class',
               default='eventlet',
               help='The type of workers to use.'),
    cfg.IntOpt('worker_connections',
               default=20,
               help='The number of concurrent worker threads.'),
    cfg.IntOpt('limit_request_line',
               default=0,
               help='The maximum size of HTTP request line in bytes.'),
    cfg.StrOpt('loglevel',
               default='debug',
               help='The granularity of Error log outputs.'),
]
cfg.CONF.register_opts(options, 'gunicorn')


LOG = logging.getLogger(__name__)


class Neutron(base.Application):
    def init(self, *args, **kwargs):
        pass

    def load_config(self):
        self.cfg = gconfig.Config(self.usage, prog=self.prog)
        settings = {'bind': '%s:%s' % (cfg.CONF.bind_host, cfg.CONF.bind_port),
                    'workers': cfg.CONF.api_workers,
                    'worker_connections': cfg.CONF.gunicorn.worker_connections,
                    'worker_class': cfg.CONF.gunicorn.worker_class,
                    'proc_name': 'neutron-server',
                    'accesslog': cfg.CONF.gunicorn.access_log,
                    'errorlog': cfg.CONF.gunicorn.error_log,
                    'limit_request_line': cfg.CONF.gunicorn.limit_request_line,
                    'loglevel': cfg.CONF.gunicorn.loglevel,
                    'access_log_format': ' '.join(('%(h)s',
                                                   '%(l)s',
                                                   '%(u)s',
                                                   '%(t)s',
                                                   '"%(r)s"',
                                                   '%(s)s',
                                                   '%(b)s',
                                                   '"%(f)s"',
                                                   '"%(a)s"',
                                                   '%(T)s',
                                                   '%(D)s',)),
                    }

        for k, v in settings.iteritems():
            self.cfg.set(k.lower(), v)

    def load(self):
        return config.load_paste_app(app_name=self.prog)

    def run(self):
        base.Arbiter(self).run()


def main():
    config.init(sys.argv[1:])
    if not cfg.CONF.config_file:
        sys.exit(_("ERROR: Unable to find configuration file via the default"
                   " search paths (~/.neutron/, ~/, /etc/neutron/, /etc/) and"
                   " the '--config-file' option!"))

    config.setup_logging()
    utils.log_opt_values(LOG)

    neutron_api = Neutron(prog='neutron')
    neutron_api.run()


if __name__ == '__main__':
    main()



"""quark_ports.created_at index for pagination sorting

Revision ID: 3a47813ce501
Revises: 4fc07b41d45c
Create Date: 2015-01-14 16:15:10.938745

"""

# revision identifiers, used by Alembic.
revision = '3a47813ce501'
down_revision = '4fc07b41d45c'

from alembic import op


def upgrade():
    op.create_index(op.f('ix_quark_ports_created_at'),
                    'quark_ports',
                    ['created_at'],
                    unique=False)


def downgrade():
        op.drop_index(op.f('ix_quark_ports_created_at'),
                      table_name='quark_ports')



# Copyright 2015 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as n_exc

from quark.db import models
from quark import tags
from quark.tests import test_base


class FakeTag(tags.Tag):

    NAME = "fake_tag"

    def validate(self, value):
        if value == "invalid":
            raise tags.TagValidationError(value, "FakeTag value is 'invalid'.")
        return True


class FooTag(tags.Tag):

    NAME = "foo_tag"

    def validate(self, value):
        if value != "foo":
            raise tags.TagValidationError(value, "FooTag value is not 'foo'.")
        return True


class FakeTagRegistry(tags.PortTagRegistry):
    def __init__(self, tags=None):
        if tags:
            self.tags = tags


class TestTagBase(test_base.TestBase):

    def setUp(self, tag=None, value=None, value2=None, invalid_value=None):
        super(TestTagBase, self).setUp()
        self.tag = tag if tag else FakeTag()
        self.value = value if value else "first"
        self.value2 = value2 if value2 else "second"
        self.invalid_value = invalid_value if invalid_value else "invalid"

        self.foo_tag = FooTag()
        self.existing_tags = ["EXISTING_TAG:already exists",
                              "random tag",
                              "123"]
        tags = {
            self.tag.get_name(): self.tag,
            self.foo_tag.get_name(): self.foo_tag
        }
        self.registry = FakeTagRegistry(tags=tags)

    def _create_test_model(self, id, tags=None):
        tags = tags if tags else []
        tags = self.existing_tags + tags
        model = models.Port(id=id, network_id="1", ip_addresses=[],
                            tags=tags)
        return model

    def _assert_tags(self, model, tags=None):
        """Assert given tags and already existing tags are present."""
        tags = tags if tags else []
        expected_tags = (self.existing_tags + tags)
        self.assertEqual(sorted(model.tags),
                         sorted(expected_tags))

    def test_tag_registry_get_all(self):
        model = self._create_test_model(1, tags=[])
        self.tag.set(model, self.value)
        self.foo_tag.set(model, "foo")
        expected_tags = [
            self.tag.serialize(self.value),
            self.foo_tag.serialize("foo")
        ]
        self._assert_tags(model, tags=expected_tags)

        tags = self.registry.get_all(model)
        self.assertEqual(tags, {self.tag.get_name(): str(self.value),
                                'foo_tag': 'foo'})

    def test_tag_registry_set_all(self):
        model = self._create_test_model(1, tags=[])
        self._assert_tags(model, tags=[])

        kwargs = {self.foo_tag.get_name(): "foo",
                  self.tag.get_name(): self.value}
        self.registry.set_all(model, **kwargs)

        expected_tags = [
            self.tag.serialize(self.value),
            self.foo_tag.serialize("foo")
        ]
        self._assert_tags(model, tags=expected_tags)

    def test_tag_registry_set_all_invalid_raises(self):
        model = self._create_test_model(1, tags=[])
        self._assert_tags(model, tags=[])

        kwargs = {self.foo_tag.get_name(): "foo",
                  self.tag.get_name(): self.invalid_value}

        with self.assertRaises(n_exc.BadRequest):
            self.registry.set_all(model, **kwargs)

    def test_tag_get(self):
        tags = [
            self.tag.serialize(self.value)
        ]
        model = self._create_test_model(1, tags=tags)
        self._assert_tags(model, tags=tags)

        self.assertEqual(self.tag.get(model), str(self.value))
        self._assert_tags(model, tags=tags)

        self.assertEqual(self.tag.get(model), str(self.value))
        self._assert_tags(model, tags=tags)

    def test_tag_get_invalid(self):
        tags = [
            self.foo_tag.serialize(self.invalid_value)
        ]
        model = self._create_test_model(1, tags=tags)
        self._assert_tags(model, tags=tags)

        self.assertEqual(self.tag.get(model), None)
        self._assert_tags(model, tags=tags)

    def test_tag_set(self):
        model = self._create_test_model(1, tags=[])
        self._assert_tags(model, tags=[])

        expected_tags = [
            self.tag.serialize(self.value)
        ]

        self.tag.set(model, self.value)
        self._assert_tags(
            model, tags=expected_tags)

        self.tag.set(model, self.value)
        self._assert_tags(
            model, tags=expected_tags)

    def test_tag_set_existing(self):
        tags = [
            self.tag.serialize(self.value)
        ]
        model = self._create_test_model(1, tags=tags)
        self._assert_tags(model, tags=tags)

        self.tag.set(model, self.value2)
        self._assert_tags(
            model, tags=[self.tag.serialize(self.value2)])

    def test_tag_set_invalid(self):
        model = self._create_test_model(1, tags=[])

        with self.assertRaises(tags.TagValidationError):
            self.tag.set(model, self.invalid_value)

        self._assert_tags(model, tags=[])

    def test_pop(self):
        tags = [
            self.tag.serialize(self.value)
        ]
        model = self._create_test_model(1, tags=tags)
        self._assert_tags(model, tags=tags)

        self.assertEqual(self.tag.pop(model), str(self.value))
        self._assert_tags(model, tags=[])

        self.assertEqual(self.tag.pop(model), None)
        self._assert_tags(model, tags=[])

    def test_pop_invalid(self):
        tags = [
            self.tag.serialize(self.invalid_value),
            self.tag.serialize(self.value),
            self.tag.serialize(self.value2)
        ]
        model = self._create_test_model(1, tags=tags)
        self._assert_tags(model, tags=tags)

        self.assertTrue(self.tag.pop(model) in
                        [str(self.value), str(self.value2)])
        self._assert_tags(model, tags=[])

        self.assertEqual(self.tag.pop(model), None)
        self._assert_tags(model, tags=[])


class TestVlanTag(TestTagBase):

    def setUp(self):
        tag = tags.VlanTag()
        value = 50
        value2 = 100
        invalid_value = 5000
        super(TestVlanTag, self).setUp(
            tag=tag, value=value, value2=value2, invalid_value=invalid_value)

    def test_vlan_validation(self):
        model = self._create_test_model(1, tags=[])

        with self.assertRaises(tags.TagValidationError):
            self.tag.set(model, self.tag.MIN_VLAN_ID - 1)
        self._assert_tags(model, tags=[])

        with self.assertRaises(tags.TagValidationError):
            self.tag.set(model, self.tag.MAX_VLAN_ID + 1)
        self._assert_tags(model, tags=[])

        with self.assertRaises(tags.TagValidationError):
            self.tag.set(model, 'three')
        self._assert_tags(model, tags=[])

        self.tag.set(model, self.tag.MIN_VLAN_ID)
        self._assert_tags(
            model, tags=[self.tag.serialize(self.tag.MIN_VLAN_ID)])

        self.tag.set(model, self.tag.MAX_VLAN_ID)
        self._assert_tags(
            model, tags=[self.tag.serialize(self.tag.MAX_VLAN_ID)])

        self.tag.set(model, str(self.tag.MIN_VLAN_ID))
        self._assert_tags(
            model, tags=[self.tag.serialize(self.tag.MIN_VLAN_ID)])

        self.tag.set(model, str(self.tag.MAX_VLAN_ID))
        self._assert_tags(
            model, tags=[self.tag.serialize(self.tag.MAX_VLAN_ID)])



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron.common import exceptions as neutron_exc
from neutron import quota
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from quark import allocation_pool
from quark.db import api as db_api
from quark import exceptions as q_exc
from quark import plugin_views as v

CONF = cfg.CONF
DEFAULT_ROUTE = netaddr.IPNetwork("0.0.0.0/0")
LOG = logging.getLogger(__name__)

ipam_driver = (importutils.import_class(CONF.QUARK.ipam_driver))()


def get_route(context, id):
    LOG.info("get_route %s for tenant %s" % (id, context.tenant_id))
    route = db_api.route_find(context, id=id, scope=db_api.ONE)
    if not route:
        raise q_exc.RouteNotFound(route_id=id)
    return v._make_route_dict(route)


def get_routes(context):
    LOG.info("get_routes for tenant %s" % context.tenant_id)
    routes = db_api.route_find(context)
    return [v._make_route_dict(r) for r in routes]


def create_route(context, route):
    LOG.info("create_route for tenant %s" % context.tenant_id)
    if not route:
        raise n_exc.BadRequest(resource="routes", msg="Malformed body")
    route = route.get("route")
    if not route:
        raise n_exc.BadRequest(resource="routes", msg="Malformed body")
    for key in ["gateway", "cidr", "subnet_id"]:
        if key not in route:
            raise n_exc.BadRequest(resource="routes",
                                   msg="%s is required" % key)

    subnet_id = route["subnet_id"]
    with context.session.begin():
        subnet = db_api.subnet_find(context, id=subnet_id, scope=db_api.ONE)
        if not subnet:
            raise n_exc.SubnetNotFound(subnet_id=subnet_id)

        if subnet["ip_policy"]:
            policies = subnet["ip_policy"].get_cidrs_ip_set()
        else:
            policies = netaddr.IPSet([])

        alloc_pools = allocation_pool.AllocationPools(subnet["cidr"],
                                                      policies=policies)
        try:
            alloc_pools.validate_gateway_excluded(route["gateway"])
        except neutron_exc.GatewayConflictWithAllocationPools as e:
            LOG.exception(str(e))
            raise n_exc.BadRequest(resource="routes", msg=str(e))

        # TODO(anyone): May want to denormalize the cidr values into columns
        #               to achieve single db lookup on conflict check
        route_cidr = netaddr.IPNetwork(route["cidr"])
        subnet_routes = db_api.route_find(context, subnet_id=subnet_id,
                                          scope=db_api.ALL)

        quota.QUOTAS.limit_check(context, context.tenant_id,
                                 routes_per_subnet=len(subnet_routes) + 1)

        for sub_route in subnet_routes:
            sub_route_cidr = netaddr.IPNetwork(sub_route["cidr"])
            if sub_route_cidr.value == DEFAULT_ROUTE.value:
                continue
            if route_cidr in sub_route_cidr or sub_route_cidr in route_cidr:
                raise q_exc.RouteConflict(route_id=sub_route["id"],
                                          cidr=str(route_cidr))
        new_route = db_api.route_create(context, **route)
    return v._make_route_dict(new_route)


def delete_route(context, id):
    # TODO(mdietz): This is probably where we check to see that someone is
    #              admin and only filter on tenant if they aren't. Correct
    #              for all the above later
    LOG.info("delete_route %s for tenant %s" % (id, context.tenant_id))
    with context.session.begin():
        route = db_api.route_find(context, id=id, scope=db_api.ONE)
        if not route:
            raise q_exc.RouteNotFound(route_id=id)
        db_api.route_delete(context, route)



"""add subnet indexes and defer

Revision ID: 80419263930a
Revises: 1bd7cff90384
Create Date: 2016-02-09 19:26:27.899610

"""

# revision identifiers, used by Alembic.
revision = '80419263930a'
down_revision = '1bd7cff90384'

from alembic import op


def upgrade():
    op.create_index(op.f('ix_quark_subnets_segement_network_version'),
                    'quark_subnets', ['segment_id', 'network_id',
                                      'ip_version', 'do_not_use'],
                    unique=False)


def downgrade():
    op.drop_index(op.f('ix_quark_subnets_segement_network_version'),
                  table_name='quark_subnets')



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import sys
import time

from neutron.common import config
from neutron.common import utils as n_utils
from oslo_config import cfg
from oslo_log import log as logging

from quark.agent import xapi
from quark.cache import security_groups_client as sg_cli

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

agent_opts = [
    cfg.IntOpt("polling_interval",
               default=10,
               help=_("Number of seconds to wait between poll iterations of "
                      "XAPI and the configured security groups registry."))
]

CONF.register_opts(agent_opts, "AGENT")


def _sleep():
    # NOTE(amir): add randomness to polling so all machines don't slam
    #             the security groups registry at once
    time.sleep(CONF.AGENT.polling_interval + random.random() * 2)


def is_isonet_vif(vif):
    """Determine if a vif is on isonet

    Returns True if a vif belongs to an isolated network by checking
    for a nicira interface id.
    """
    nicira_iface_id = vif.record.get('other_config').get('nicira-iface-id')

    if nicira_iface_id:
        return True

    return False


def partition_vifs(xapi_client, interfaces, security_group_states):
    """Splits VIFs into three explicit categories and one implicit

    Added - Groups exist in Redis that have not been ack'd and the VIF
            is not tagged.
            Action: Tag the VIF and apply flows
    Updated - Groups exist in Redis that have not been ack'd and the VIF
              is already tagged
              Action: Do not tag the VIF, do apply flows
    Removed - Groups do NOT exist in Redis but the VIF is tagged
              Action: Untag the VIF, apply default flows
    Self-Heal - Groups are ack'd in Redis but the VIF is untagged. We treat
                this case as if it were an "added" group.
                Action: Tag the VIF and apply flows
    NOOP - The VIF is not tagged and there are no matching groups in Redis.
           This is our implicit category
           Action: Do nothing
    """
    added = []
    updated = []
    removed = []

    for vif in interfaces:
        # Quark should not action on isonet vifs in regions that use FLIP
        if ('floating_ip' in CONF.QUARK.environment_capabilities and
                is_isonet_vif(vif)):
            continue

        vif_has_groups = vif in security_group_states
        if vif.tagged and vif_has_groups and security_group_states[vif]:
            # Already ack'd these groups and VIF is tagged, reapply.
            # If it's not tagged, fall through and have it self-heal
            continue

        if vif.tagged:
            if vif_has_groups:
                updated.append(vif)
            else:
                removed.append(vif)
        else:
            if vif_has_groups:
                added.append(vif)
            # if not tagged and no groups, skip

    return added, updated, removed


def ack_groups(client, groups):
    if len(groups) > 0:
        client.update_group_states_for_vifs(groups, True)


def run():
    """Fetches changes and applies them to VIFs periodically

    Process as of RM11449:
    * Get all groups from redis
    * Fetch ALL VIFs from Xen
    * Walk ALL VIFs and partition them into added, updated and removed
    * Walk the final "modified" VIFs list and apply flows to each
    """
    groups_client = sg_cli.SecurityGroupsClient()
    xapi_client = xapi.XapiClient()

    interfaces = set()
    while True:
        try:
            interfaces = xapi_client.get_interfaces()
        except Exception:
            LOG.exception("Unable to get instances/interfaces from xapi")
            _sleep()
            continue

        try:
            sg_states = groups_client.get_security_group_states(interfaces)
            new_sg, updated_sg, removed_sg = partition_vifs(xapi_client,
                                                            interfaces,
                                                            sg_states)
            xapi_client.update_interfaces(new_sg, updated_sg, removed_sg)
            groups_to_ack = [v for v in new_sg + updated_sg if v.success]
            ack_groups(groups_client, groups_to_ack)

        except Exception:
            LOG.exception("Unable to get security groups from registry and "
                          "apply them to xapi")
            _sleep()
            continue

        _sleep()


def main():
    config.init(sys.argv[1:])
    config.setup_logging()
    n_utils.log_opt_values(LOG)
    if not CONF.config_file:
        sys.exit(_("ERROR: Unable to find configuration file via the default"
                   " search paths (~/.neutron/, ~/, /etc/neutron/, /etc/) and"
                   " the '--config-file' option!"))
    run()



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import mock
import netaddr

from quark.db import api as db_api
import quark.ipam
import quark.plugin
import quark.plugin_modules.mac_address_ranges as macrng_api
import quark.plugin_modules.networks as network_api
import quark.plugin_modules.ports as port_api
import quark.plugin_modules.subnets as subnet_api
from quark.tests.functional.base import BaseFunctionalTest


class QuarkNetworksPaginationFunctionalTest(BaseFunctionalTest):
    def test_networks_pagination(self):
        networks_per_page = 1
        networkA = dict(name="fake_A", tenant_id="fake", network_plugin="BASE")
        networkB = dict(name="fake_B", tenant_id="fake", network_plugin="BASE")
        db_api.network_create(self.context, **networkA)
        db_api.network_create(self.context, **networkB)
        res = network_api.get_networks(self.context, networks_per_page,
                                       [('id', 'asc')], None, False,
                                       None)
        self.assertEqual(len(res), networks_per_page)
        res = network_api.get_networks(self.context)
        self.assertNotEqual(len(res), networks_per_page)


class QuarkSubnetsPaginationFunctionalTest(BaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnets):
        self.ipam = quark.ipam.QuarkIpamANY()
        with self.context.session.begin():
            net_mod = db_api.network_create(self.context, **network)
            for subnet in subnets:
                subnet["network"] = net_mod
                db_api.subnet_create(self.context, **subnet)
        yield net_mod

    def test_subnet_pagination(self):
        subnets_per_page = 1
        network = dict(name="public", tenant_id="fake")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 1
        subnet_1 = dict(id=1, cidr="0.0.0.0/24", next_auto_assign_ip=next_ip,
                        ip_policy=None, tenant_id="fake")
        subnet_2 = dict(id=2, cidr="1.1.1.0/24", next_auto_assign_ip=next_ip,
                        ip_policy=None, tenant_id="fake")
        with self._stubs(network, [subnet_1, subnet_2]):
            subnets_unpaged = subnet_api.get_subnets(self.context, filters={})
            subnets_paged = subnet_api.get_subnets(self.context,
                                                   subnets_per_page, False,
                                                   [('id', 'asc')],
                                                   filters={})
            self.assertEqual(len(subnets_paged), subnets_per_page)
            self.assertNotEqual(len(subnets_unpaged), subnets_per_page)


class QuarkPortsPaginationFunctionalTest(BaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_info):
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            self.context.is_admin = True
            net = network_api.create_network(self.context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            sub_ports = []
            subnet_info['subnet']['network_id'] = net['id']
            sub_ports.append(subnet_api.create_subnet(self.context,
                                                      subnet_info))
            yield net, sub_ports

    def test_ports_pagination(self):
        cidr = "192.168.1.0/24"

        ip_network = netaddr.IPNetwork(cidr)

        network = dict(id="1", name="public", tenant_id="make",
                       network_plugin="BASE",
                       ipam_strategy="ANY")
        network = {"network": network}

        subnet = dict(id="1", ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake")
        subnet_info = {"subnet": subnet}

        def _make_body():
            fix_ipv4 = dict(subnet_id=sub_ports[0]['id'])
            port_info = {"port": dict(fixed_ips=[fix_ipv4],
                                      network_id=net['id'])}
            return port_info

        ports_per_page = 1
        with self._stubs(network, subnet_info) as (
                net, sub_ports):
            port_api.create_port(self.context, _make_body())
            port_api.create_port(self.context, _make_body())
            res_ports = port_api.get_ports(self.context, ports_per_page,
                                           [('id', 'asc')], None)
            self.assertEqual(len(res_ports), ports_per_page)
            res_ports = port_api.get_ports(self.context)
            self.assertNotEqual(len(res_ports), ports_per_page)
            # Note (Perkins): Testing for a default sort on created_at,
            # but created_at is not available, so check that mac addresses,
            # which are created sequentially, are ordered correctly.
            res_ports = port_api.get_ports(self.context, 2, None, None)
            self.assertTrue(res_ports[0]['mac_address'] <
                            res_ports[1]['mac_address'])
